import pickle
import re
import time

import numpy as np
import torch
from tqdm import tqdm
import torch.nn.functional as F

from plants_sm.data_structures.dataset import PLACEHOLDER_FIELD
from plants_sm.io.pickle import write_pickle


def get_ec_from_regex_match(match):
    if match is not None:
        EC = match.group()
        if EC is not None:
            return EC
    return None


def divide_labels_by_EC_level(ec_numbers):
    ECs = ec_numbers.split(";")
    # get the first 3 ECs with regular expression
    EC3 = []
    EC2 = []
    EC1 = []
    EC4 = []
    for EC in ECs:
        new_EC = re.search(r"^\d+.\d+.\d+.n*\d+", EC)
        new_EC = get_ec_from_regex_match(new_EC)
        if isinstance(new_EC, str):
            if new_EC not in EC4:
                EC4.append(new_EC)

        new_EC = re.search(r"^\d+.\d+.\d+", EC)
        new_EC = get_ec_from_regex_match(new_EC)
        if isinstance(new_EC, str):
            if new_EC not in EC3:
                EC3.append(new_EC)

        new_EC = re.search(r"^\d+.\d+", EC)
        new_EC = get_ec_from_regex_match(new_EC)
        if isinstance(new_EC, str):
            if new_EC not in EC2:
                EC2.append(new_EC)

        new_EC = re.search(r"^\d+", EC)
        new_EC = get_ec_from_regex_match(new_EC)
        if isinstance(new_EC, str):
            if new_EC not in EC1:
                EC1.append(new_EC)

    return EC1, EC2, EC3, EC4


def get_ec_id_dict(dataset, EC_label) -> tuple:
    id_ec = {}
    ec_id = {}
    labels_to_consider = dataset._labels_names

    for i, rows in dataset.dataframe.iterrows():
        sequence_id = rows[dataset.instances_ids_field]
        EC = rows[EC_label]
        if i > 0:
            EC1, EC2, EC3, EC4 = divide_labels_by_EC_level(EC)
            all_ecs = EC1 + EC2 + EC3 + EC4
            for ec in all_ecs:
                if ec not in labels_to_consider:
                    continue
                else:
                    if sequence_id not in id_ec.keys():
                        id_ec[sequence_id] = []
                        id_ec[sequence_id].append(ec)
                    else:
                        id_ec[sequence_id].append(ec)
            if ec in labels_to_consider:
                for ec in id_ec[sequence_id]:
                    if ec not in ec_id.keys():
                        ec_id[ec] = set()
                        ec_id[ec].add(sequence_id)
                    else:
                        ec_id[ec].add(sequence_id)
    return id_ec, ec_id


def esm_embedding(dataset, ec_id_dict, device, dtype):
    '''
    Loading esm embedding in the sequence of EC numbers
    prepare for calculating cluster center by EC
    '''
    esm_emb = []
    # for ec in tqdm(list(ec_id_dict.keys())):
    for ec in list(ec_id_dict.keys()):
        ids_for_query = list(ec_id_dict[ec])
        esm_to_cat = [torch.from_numpy(dataset.features[PLACEHOLDER_FIELD][id_]).unsqueeze(0) for id_ in ids_for_query]
        esm_emb = esm_emb + esm_to_cat
    return torch.cat(esm_emb).to(device=device, dtype=dtype)


def model_embedding_test(dataset, id_ec_test, model, device, dtype):
    '''
    Instead of loading esm embedding in the sequence of EC numbers
    the test embedding is loaded in the sequence of queries
    then inferenced with model to get model embedding
    '''
    ids_for_query = list(id_ec_test.keys())
    esm_to_cat = [torch.from_numpy(dataset.features[PLACEHOLDER_FIELD][id_]).unsqueeze(0)
                  for id_ in ids_for_query]
    esm_emb = torch.cat(esm_to_cat).to(device=device, dtype=dtype)
    model_emb = model(esm_emb)
    return model_emb


def model_embedding_test_ensemble(dataset, id_ec_test, device, dtype):
    '''
    Instead of loading esm embedding in the sequence of EC numbers
    the test embedding is loaded in the sequence of queries
    '''
    ids_for_query = list(id_ec_test.keys())
    esm_to_cat = [torch.from_numpy(dataset.features[PLACEHOLDER_FIELD][id_]).unsqueeze(0)
                  for id_ in ids_for_query]
    esm_emb = torch.cat(esm_to_cat).to(device=device, dtype=dtype)
    return esm_emb


def get_dist_map_test(model_emb_train, model_emb_test,
                      ec_id_dict_train, id_ec_test,
                      device, dtype, dot=False):
    '''
    Get the pair-wise distance map for test queries and train EC cluster centers
    map is of size of (N_test_ids, N_EC_train)
    '''
    print("The embedding sizes for train and test:",
          model_emb_train.size(), model_emb_test.size())
    # get cluster center for all EC appeared in training set
    cluster_center_model = get_cluster_center(
        model_emb_train, ec_id_dict_train)
    total_ec_n, out_dim = len(ec_id_dict_train.keys()), model_emb_train.size(1)
    model_lookup = torch.zeros(total_ec_n, out_dim, device=device, dtype=dtype)
    ecs = list(cluster_center_model.keys())
    for i, ec in enumerate(ecs):
        model_lookup[i] = cluster_center_model[ec]
    model_lookup = model_lookup.to(device=device, dtype=dtype)
    # calculate distance map between n_query_test * total_ec_n (training) pairs
    ids = list(id_ec_test.keys())
    print(f'Calculating eval distance map, between {len(ids)} test ids '
          f'and {total_ec_n} train EC cluster centers')
    if dot:
        eval_dist = dist_map_helper_dot(ids, model_emb_test, ecs, model_lookup)
    else:
        eval_dist = dist_map_helper(ids, model_emb_test, ecs, model_lookup)
    return eval_dist


def get_random_nk_dist_map(emb_train, rand_nk_emb_train,
                           ec_id_dict_train, rand_nk_ids,
                           device, dtype, dot=False):
    '''
    Get the pair-wise distance map between
    randomly chosen nk ids from training and all EC cluster centers
    map is of size of (nk, N_EC_train)
    '''

    cluster_center_model = get_cluster_center(emb_train, ec_id_dict_train)
    total_ec_n, out_dim = len(ec_id_dict_train.keys()), emb_train.size(1)
    model_lookup = torch.zeros(total_ec_n, out_dim, device=device, dtype=dtype)
    ecs = list(cluster_center_model.keys())
    for i, ec in enumerate(ecs):
        model_lookup[i] = cluster_center_model[ec]
    model_lookup = model_lookup.to(device=device, dtype=dtype)
    if dot:
        random_nk_dist_map = dist_map_helper_dot(
            rand_nk_ids, rand_nk_emb_train, ecs, model_lookup)
    else:
        random_nk_dist_map = dist_map_helper(
            rand_nk_ids, rand_nk_emb_train, ecs, model_lookup)
    return random_nk_dist_map


def random_nk_model(id_ec_train, ec_id_dict_train, emb_train, n=10, weighted=False):
    ids = list(id_ec_train.keys())
    nk = n * 1000
    if weighted:
        P = []
        for id in id_ec_train.keys():
            ecs_id = id_ec_train[id]
            ec_densities = [len(ec_id_dict_train[ec]) for ec in ecs_id]
            # the prob of calling this id is inversely prop to 1/max(density)
            P.append(1 / np.max(ec_densities))
        P = P / np.sum(P)
        random_nk_id = np.random.choice(
            range(len(ids)), nk, replace=True, p=P)
    else:
        random_nk_id = np.random.choice(range(len(ids)), nk, replace=False)

    random_nk_id = np.sort(random_nk_id)
    chosen_ids = [ids[i] for i in random_nk_id]
    chosen_emb_train = emb_train[random_nk_id]
    return chosen_ids, chosen_emb_train


def get_cluster_center(model_emb, ec_id_dict):
    cluster_center_model = {}
    id_counter = 0
    with torch.no_grad():
        for ec in tqdm(list(ec_id_dict.keys())):
            ids_for_query = list(ec_id_dict[ec])
            id_counter_prime = id_counter + len(ids_for_query)
            emb_cluster = model_emb[id_counter: id_counter_prime]
            cluster_center = emb_cluster.mean(dim=0)
            cluster_center_model[ec] = cluster_center.detach().cpu()
            id_counter = id_counter_prime
    return cluster_center_model


def dist_map_helper(keys1, lookup1, keys2, lookup2):
    dist = {}
    bar = tqdm(total=len(keys1))
    for i, key1 in enumerate(keys1):
        current = lookup1[i].unsqueeze(0)
        dist_norm = (current - lookup2).norm(dim=1, p=2)
        dist_norm = dist_norm.detach().cpu().numpy()
        dist[key1] = {}
        bar.update(1)
        for j, key2 in enumerate(keys2):
            dist[key1][key2] = dist_norm[j]
    return dist



def get_dist_map(ec_id_dict, esm_emb, device, dtype, model=None, dot=False):
    '''
    Get the distance map for training, size of (N_EC_train, N_EC_train)
    between all possible pairs of EC cluster centers
    '''
    # inference all queries at once to get model embedding
    if model is not None:
        model_emb = model(esm_emb.to(device=device, dtype=dtype))
    else:
        # the first distance map before training comes from ESM
        model_emb = esm_emb
    # calculate cluster center by averaging all embeddings in one EC
    cluster_center_model = get_cluster_center(model_emb, ec_id_dict)
    # organize cluster centers in a matrix
    total_ec_n, out_dim = len(ec_id_dict.keys()), model_emb.size(1)
    model_lookup = torch.zeros(total_ec_n, out_dim, device=device, dtype=dtype)
    ecs = list(cluster_center_model.keys())
    for i, ec in enumerate(ecs):
        model_lookup[i] = cluster_center_model[ec]
    model_lookup = model_lookup.to(device=device, dtype=dtype)
    # calculate pairwise distance map between total_ec_n * total_ec_n pairs
    print(f'Calculating distance map, number of unique EC is {total_ec_n}')
    if dot:
        model_dist = dist_map_helper_dot(ecs, model_lookup, ecs, model_lookup)
    else:
        model_dist = dist_map_helper(ecs, model_lookup, ecs, model_lookup)
    return model_dist


def dist_map_helper_dot(keys1, lookup1, keys2, lookup2):
    dist = {}
    lookup1 = F.normalize(lookup1, dim=-1, p=2)
    lookup2 = F.normalize(lookup2, dim=-1, p=2)
    for i, key1 in tqdm(enumerate(keys1)):
        current = lookup1[i].unsqueeze(0)
        dist_norm = (current - lookup2).norm(dim=1, p=2)
        dist_norm = dist_norm ** 2
        # dist_norm = (current - lookup2).norm(dim=1, p=2)
        dist_norm = dist_norm.detach().cpu().numpy()
        dist[key1] = {}
        for j, key2 in enumerate(keys2):
            dist[key1][key2] = dist_norm[j]
    return dist


def compute_esm_distance(dataset, distance_map_path, device):
    _, ec_id_dict = get_ec_id_dict(dataset, "EC")
    dtype = torch.float32
    esm_emb = esm_embedding(dataset, ec_id_dict, device, dtype)
    esm_dist = get_dist_map(ec_id_dict, esm_emb, device, dtype)
    write_pickle(distance_map_path + '.pkl', esm_dist)
    write_pickle(distance_map_path + '_esm.pkl', esm_emb)



import logging
import os
import shutil
from pathlib import Path
from typing import Union, Optional, Dict
from urllib import request

import torch
import yaml
from numpy import ndarray
from torch._appdirs import user_cache_dir
from yaml import YAMLError

from plants_sm.io.yaml import read_yaml

logger = logging.getLogger(__name__)

_module_dir: Path = Path(os.path.dirname(os.path.abspath(__file__)))


def get_device(device: Union[None, str, torch.device] = None, cuda_as_default=False) -> torch.device:
    """
    Returns what the user specified, or defaults to the GPU,
    with a fallback to CPU if no GPU is available.

    Parameters
    ----------
    device: Union[None, str, torch.device]
        The device to use. If None, defaults to GPU, with fallback to CPU.
    cuda_as_default: bool
        Whether to use CUDA as default device.
    Returns
    -------
    torch.device
        The device to use.
    """
    if isinstance(device, torch.device):
        return device
    elif device is None:
        return torch.device("cpu")
    elif "cuda" in device:
        is_available = torch.cuda.is_available()
        if is_available:
            return torch.device(device)
        else:
            logger.warning("CUDA is not available, falling back to CPU.")
            return torch.device("cpu")
    elif torch.cuda.is_available() and cuda_as_default:
        return torch.device("cuda")
    else:
        return torch.device("cpu")


def read_config_file(config_path: Union[str, Path], preserve_order: bool = True) -> dict:
    """
    Read config from path to file.

    Parameters
    ----------
    config_path: Union[str, Path]
        Path to config file.
    preserve_order: bool
        Whether to preserve order of config file.

    Returns
    -------
    result of parsing the YAML file: dict
    """
    read_yaml(config_path, preserve_order=preserve_order)
    with open(config_path, "r") as fp:
        try:
            if preserve_order:
                return yaml.load(fp, Loader=yaml.Loader)
            else:
                return yaml.safe_load(fp)
        except YAMLError as e:
            raise YAMLError(
                f"Could not parse configuration file at '{config_path}' as yaml. "
                "Formatting mistake in config file? "
                "See Error above for details."
            ) from e


def get_model_file(model: Optional[str] = None,
                   file: Optional[str] = None,
                   overwrite_cache: bool = False) -> str:
    """
    If the specified asset for the model is in the user cache, returns the
    location, otherwise downloads the file to cache and returns the location.

    Parameters
    ----------
    model: Optional[str]
        The model to get the file for. If None, returns the file for the
        default model.
    file: Optional[str]
        The file to get. If None, returns the default file for the model.
    overwrite_cache: bool
        Whether to overwrite the cache if the file is already present.
    """
    cache_path = Path(user_cache_dir("plants_sm")).joinpath(model).joinpath(file)
    if not overwrite_cache and cache_path.is_file():
        logger.info(f"Loading {file} for {model} from cache at '{cache_path}'")
        return str(cache_path)

    cache_path.parent.mkdir(exist_ok=True, parents=True)
    _defaults: Dict[str, Dict[str, str]] = read_config_file(_module_dir / "defaults.yml")
    url = _defaults.get(model, {}).get(file)

    # Since the files are not user provided, this must never happen
    assert url, f"File {file} for {model} doesn't exist."

    logger.info(f"Downloading {file} for {model} and storing it in '{cache_path}'")

    req = request.Request(url, headers={
        'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64)'
    })

    with request.urlopen(req) as response, open(cache_path, 'wb') as outfile:
        shutil.copyfileobj(response, outfile)

    return str(cache_path)


def reduce_per_protein(embedding: ndarray) -> ndarray:
    """
    Reduce the embedding of a protein to a single vector.

    Parameters
    ----------
    embedding: ndarray
        The embedding of a protein.

    Returns
    -------
    ndarray
        The reduced embedding.
    """
    # This is `h_avg` in jax-unirep terminology
    return embedding.mean(axis=0)



import sys

import nni

from plants_sm.models.enumerators import ModelFileEnumeratorsUtils

if __name__ == "__main__":
    model_path = sys.argv[0]
    RECEIVED_PARAMS = nni.get_next_parameter()
    PARAMS = {}
    PARAMS.update(RECEIVED_PARAMS)
    model_type = ModelFileEnumeratorsUtils.get_class_to_load_from_directory(model_path)
    loaded_model = model_type.load(model_path)
    loaded_model.mode.__init__(PARAMS)
    loaded_model.fit()



from typing import Dict, Tuple

from rdkit.Chem import MolFromSmiles, MolToSmiles

from plants_sm.data_standardization.compounds._presets import DEEPMOL_STANDARDIZERS
from plants_sm.data_structures.dataset import Dataset
from plants_sm.transformation._utils import transform_instances
from plants_sm.transformation.transformer import Transformer


class DeepMolStandardizer(Transformer):
    preset: str = "custom_standardizer"
    kwargs: Dict = {}

    def _fit(self, dataset: Dataset, instance_type: str) -> 'DeepMolStandardizer':
        """
        Method to fit the transformer

        Parameters
        ----------

        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        self: DeepMolStandardizer
        """
        if self.preset not in DEEPMOL_STANDARDIZERS:
            raise ValueError(f'Preset {self.preset} is not available.')

        descriptor = DEEPMOL_STANDARDIZERS[self.preset]
        self.descriptor = descriptor(**self.kwargs)
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'DeepMolStandardizer':
        """
        Method to fit the transformer

        Parameters
        ----------

        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        self: DeepMolStandardizer
        """
        return self._fit(dataset, instance_type)

    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        Method to transform the dataset

        Parameters
        ----------
        dataset: Dataset
            dataset to be transformed

        Returns
        -------
        dataset: Dataset
            transformed dataset

        """
        return transform_instances(self.n_jobs, dataset, self._compound_preprocessing, instance_type, self.preset)

    def _compound_preprocessing(self, compound: str, identifier: str) -> Tuple[str, str]:
        """
        Method to preprocess a compound

        Parameters
        ----------

        compound: str
            compound to be preprocessed

        identifier: str
            identifier of the compound

        Returns
        -------
        dict: Dict[str, str]
            dictionary with the identifier of the compound and the preprocessed compound
        """
        mol = MolFromSmiles(compound)
        return identifier, MolToSmiles(self.descriptor._standardize(mol))



from .psycho_chemical import (LengthDescriptor,
                              ChargeDescriptor,
                              ChargeDensityDescriptor,
                              FormulaDescriptor,
                              BondDescriptor,
                              MolecularWeightDescriptor,
                              GravyDescriptor,
                              AromacityDescriptor,
                              IsoelectricPointDescriptor,
                              InstabilityIndexDescriptor,
                              SecondaryStructureDescriptor,
                              MolarExtinctionCoefficientDescriptor,
                              FlexibilityDescriptor,
                              AliphaticIndexDescriptor,
                              BomanIndexDescriptor,
                              HydrophobicRatioDescriptor,
                              ModlAMPMomentDescriptor,
                              ModlAMPGlobalDescriptor,
                              ModlAMPProfileDescriptor,
                              ModlAMPArcDescriptor)
from .amino_acid_composition import (AminoAcidCompositionDescriptor,
                                     DipeptideCompositionDescriptor,
                                     TripeptideCompositionDescriptor)
from .pseudo_amino_acid_composition import (PseudoAminoAcidCompositionDescriptor,
                                            AmphiphilicPseudoAminoAcidCompositionDescriptor)
from .correlation import (NormalizedMoreauBrotoDescriptor,
                          MoranAutoDescriptor,
                          GearyAutoDescriptor,
                          ModlAMPAutoCorrelationDescriptor,
                          ModlAMPCrossCorrelationDescriptor)
from .composition_transition_distribution import CompositionTransitionDistributionDescriptor
from .sequence_order import (SequenceOrderCouplingNumbersDescriptor,
                             QuasiSequenceOrderDescriptor)


PHYSICO_CHEMICAL_DESCRIPTORS = [
    LengthDescriptor,
    ChargeDescriptor,
    ChargeDensityDescriptor,
    FormulaDescriptor,
    BondDescriptor,
    MolecularWeightDescriptor,
    GravyDescriptor,
    AromacityDescriptor,
    IsoelectricPointDescriptor,
    InstabilityIndexDescriptor,
    SecondaryStructureDescriptor,
    MolarExtinctionCoefficientDescriptor,
    FlexibilityDescriptor,
    AliphaticIndexDescriptor,
    BomanIndexDescriptor,
    HydrophobicRatioDescriptor
]


AMINO_ACID_COMPOSITION_DESCRIPTORS = [AminoAcidCompositionDescriptor,
                                      DipeptideCompositionDescriptor,
                                      TripeptideCompositionDescriptor]


PSEUDO_AMINO_ACID_COMPOSITION_DESCRIPTORS = [PseudoAminoAcidCompositionDescriptor,
                                             AmphiphilicPseudoAminoAcidCompositionDescriptor]


AUTO_CORRELATION_DESCRIPTORS = [NormalizedMoreauBrotoDescriptor,
                                MoranAutoDescriptor,
                                GearyAutoDescriptor]

CTD_DESCRIPTORS = [CompositionTransitionDistributionDescriptor]


SEQUENCE_ORDER_DESCRIPTORS = [SequenceOrderCouplingNumbersDescriptor,
                              QuasiSequenceOrderDescriptor]


MODLAMP_CORRELATION_DESCRIPTORS = [ModlAMPAutoCorrelationDescriptor,
                                   ModlAMPCrossCorrelationDescriptor]


MODLAM_PHYSICO_CHEMICAL_DESCRIPTORS = [ModlAMPMomentDescriptor,
                                       ModlAMPGlobalDescriptor,
                                       ModlAMPProfileDescriptor,
                                       ModlAMPArcDescriptor]


DESCRIPTORS_PRESETS = {
    'all': [*PHYSICO_CHEMICAL_DESCRIPTORS,
            *AMINO_ACID_COMPOSITION_DESCRIPTORS,
            *PSEUDO_AMINO_ACID_COMPOSITION_DESCRIPTORS,
            *AUTO_CORRELATION_DESCRIPTORS,
            *CTD_DESCRIPTORS,
            *SEQUENCE_ORDER_DESCRIPTORS,
            *MODLAMP_CORRELATION_DESCRIPTORS,
            *MODLAM_PHYSICO_CHEMICAL_DESCRIPTORS],
    'performance': [*PHYSICO_CHEMICAL_DESCRIPTORS,
                    *AMINO_ACID_COMPOSITION_DESCRIPTORS,
                    *PSEUDO_AMINO_ACID_COMPOSITION_DESCRIPTORS,
                    *CTD_DESCRIPTORS,
                    *MODLAMP_CORRELATION_DESCRIPTORS],
    'physicochemical': PHYSICO_CHEMICAL_DESCRIPTORS,
    'aac': AMINO_ACID_COMPOSITION_DESCRIPTORS,
    'paac': PSEUDO_AMINO_ACID_COMPOSITION_DESCRIPTORS,
    'auto-correlation': AUTO_CORRELATION_DESCRIPTORS,
    'composition-transition-distribution': CTD_DESCRIPTORS,
    'seq-order': SEQUENCE_ORDER_DESCRIPTORS,
    'modlamp-correlation': MODLAMP_CORRELATION_DESCRIPTORS,
    'modlamp-all': [*MODLAMP_CORRELATION_DESCRIPTORS,
                    *MODLAM_PHYSICO_CHEMICAL_DESCRIPTORS]
}



from typing import Dict

import numpy as np
from rdkit.Chem import MolFromSmiles

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.compounds._presets import DEEPMOL_PRESETS
from plants_sm.featurization.featurizer import FeaturesGenerator


class DeepMolDescriptors(FeaturesGenerator):

    preset: str = "morgan_fingerprints"
    kwargs: Dict = {}

    def set_features_names(self):
        """
        Method to set the names of the features
        """
        self.features_names = self.descriptor.feature_names

    def _fit(self, dataset: Dataset, instance_type: str) -> 'DeepMolDescriptors':
        """
        Method to fit the transformer

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        self: DeepMolDescriptors

        """

        if self.preset not in DEEPMOL_PRESETS:
            raise ValueError(f'Preset {self.preset} is not available.')

        descriptor = DEEPMOL_PRESETS[self.preset]
        self.descriptor = descriptor(**self.kwargs)
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'FeaturesGenerator':
        """
        Method to fit the transformer

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        instance_type: str
            type of the instance to be processed

        Returns
        -------
        self: DeepMolDescriptors

        """

        return self._fit(dataset, instance_type)

    def _featurize(self, molecule: str) -> np.ndarray:
        """
        Method to featurize a molecule

        Parameters
        ----------
        molecule: str
            SMILES string of the molecule to be featurized

        Returns
        -------
        features: np.ndarray
        """

        mol = MolFromSmiles(molecule)

        features = self.descriptor._featurize(mol)

        if not self.features_names:
            if features.shape[0] != np.NaN:
                self.features_names = [f"{self.preset}_{i}" for i in range(1, features.shape[0] + 1)]

        return features



import math

from ._utils import Descriptor
from ._constants import CTD_HYDROPHOBICITY_TABLE, CTD_NORMALIZED_VDWV_TABLE, CTD_POLARITY_TABLE, CTD_CHARGE_TABLE, \
    CTD_SECONDARY_STRUCTURE_TABLE, CTD_SOLVENT_ACCESSIBILITY_TABLE, POLARIZABILITY_TABLE


_CTD_TABLES = [CTD_HYDROPHOBICITY_TABLE, CTD_NORMALIZED_VDWV_TABLE, CTD_POLARITY_TABLE, CTD_CHARGE_TABLE,
               CTD_SECONDARY_STRUCTURE_TABLE, CTD_SOLVENT_ACCESSIBILITY_TABLE, POLARIZABILITY_TABLE]

_CTD_TABLE_NAMES = ['hydrophobicity', 'normalized_vdwv', 'polarity', 'charge', 'secondary_structure',
                    'solvent_accessibility', 'polarizability']


def calculate_composition(sequence, table):
    """
    It calculates the composition of a sequence.
    """
    n = len(sequence)
    translated_sequence = str.translate(sequence, table)
    c1 = round(float(translated_sequence.count('1')) / n, 3)
    c2 = round(float(translated_sequence.count('2')) / n, 3)
    c3 = round(float(translated_sequence.count('3')) / n, 3)
    return c1, c2, c3


def calculate_transition(sequence, table):
    """
    It calculates the transition of a sequence.
    """
    n = len(sequence)
    translated_sequence = str.translate(sequence, table)
    t12 = round(float(translated_sequence.count('12') + translated_sequence.count('21')) / (n - 1), 3)
    t13 = round(float(translated_sequence.count('13') + translated_sequence.count('31')) / (n - 1), 3)
    t23 = round(float(translated_sequence.count('23') + translated_sequence.count('32')) / (n - 1), 3)
    return t12, t13, t23


def calculate_distribution(sequence, table):
    """
    It calculates the distribution of a sequence.
    """
    n = len(sequence)
    translated_sequence = str.translate(sequence, table)

    result = []
    for i in ('1', '2', '3'):
        indexes = [j for j, char in enumerate(translated_sequence) if char == i]
        cnt = len(indexes)
        if cnt > 0:
            # it comprehends D_i_000, D_i_025, D_i_050, D_i_075, D_i_100
            to_extend = [
                round(float(indexes[0]) / n * 100, 3),
                round(float(indexes[int(math.floor(cnt * 0.25)) - 1]) / n * 100, 3),
                round(float(indexes[int(math.floor(cnt * 0.5)) - 1]) / n * 100, 3),
                round(float(indexes[int(math.floor(cnt * 0.75)) - 1]) / n * 100, 3),
                round(float(indexes[-1]) / n * 100, 3)
            ]
            result.extend(to_extend)

        else:
            # it comprehends D_i_000, D_i_025, D_i_050, D_i_075, D_i_100
            result.extend([0, 0, 0, 0, 0])

    return result


class CompositionTransitionDistributionDescriptor(Descriptor):
    """
    It calculates the composition, transition and distribution of a sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        It calculates the composition, transition and distribution of a sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the composition, transition and distribution for.

        kwargs

        Returns
        -------
        list of float
            the composition, transition and distribution of a sequence.
        """
        result = []
        for fn in (calculate_composition, calculate_transition, calculate_distribution):
            for table in _CTD_TABLES:
                result.extend(fn(sequence, table))
        return result

    def get_features_out(self, **kwargs):
        result = []
        for suffix in ('C', 'T', 'D'):

            if suffix == 'C':
                suffixes = [f'{suffix}_1', f'{suffix}_2', f'{suffix}_3']

            elif suffix == 'T':
                suffixes = [f'{suffix}_12', f'{suffix}_13', f'{suffix}_23']

            else:
                suffixes = [f'{suffix}_{i}_{j}' for i in ('1', '2', '3')
                            for j in ('000', '025', '050', '075', '100')]

            for name in _CTD_TABLE_NAMES:
                to_extend = [f'{name}_{suffix}' for suffix in suffixes]
                result.extend(to_extend)

        return result



import os
import shutil
from abc import abstractmethod
from typing import Any, Dict, Union, Iterable, List

import numpy as np
import pandas as pd
from cached_property import cached_property

from plants_sm.data_structures.dataset.batch_manager.batch_manager import BatchManager
from plants_sm.design_patterns.observer import ConcreteSubject
from plants_sm.io.pickle import write_pickle, read_pickle
from plants_sm.mixins.mixins import PickleMixin


class Dataset(ConcreteSubject, PickleMixin):
    representation_fields: Dict[str, Any]
    batch_size: Union[int, None] = None
    _dataframe_generator: Iterable = None
    _batch_state: bool = True
    variables_to_save = [
        ("_dataframe", "pkl"),
        ("_instances", "pkl"),
        ("_identifiers", "pkl"),
        ("features", "pkl"),
    ]

    def __init__(self, batch_size: Union[int, None] = None):
        """
        Class that represents a dataset.

        Parameters
        ----------
        batch_size: int
            the size of the batches
        """
        self.batch_size = batch_size
        if self.batch_size is not None:
            self._folder_to_load_features = None
            if self.batch_size <= 0:
                raise ValueError("Batch size must be a positive integer.")
            else:
                manager = BatchManager(batch_size=self.batch_size)
                manager.register_class(self, self.variables_to_save)
                self.attach(manager)

    def _clear_cached_properties(self):
        """
        Clears the cached properties of the class.
        """
        for name in dir(type(self)):
            if isinstance(getattr(type(self), name), cached_property):
                vars(self).pop(name, None)

    def reset_batch(self):
        """
        Method to reset the batch.
        """
        self._batch_state = False
        self._observer.end()

    def save_features(self, folder_path: str):
        """
        Method to save the features of the dataset.

        Parameters
        ----------
        folder_path: str
            The path to the folder where the features will be saved.
        """
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        if self.batch_size is not None:
            i = 0
            while self.next_batch():
                write_pickle(os.path.join(folder_path, f"features_{i}.pkl"), self.features)
                i += 1
        else:
            write_pickle(os.path.join(folder_path, f"features.pkl"), self.features)

    def load_features(self, folder_path: str):
        """
        Method to load the features of the dataset.

        Parameters
        ----------
        folder_path: str
            The path to the file where the features are saved.
        """
        if self.batch_size is not None:
            self._folder_to_load_features = folder_path
            i = 0
            self.features = read_pickle(os.path.join(folder_path, f"features_{i}.pkl"))
            i += 1
            while self.next_batch():
                self.features = read_pickle(os.path.join(folder_path, f"features_{i}.pkl"))
                i += 1
        else:
            self.features = read_pickle(os.path.join(folder_path, f"features.pkl"))

    def save(self, folder_path: str):
        """
        Method to save the dataset.

        Parameters
        ----------
        folder_path: str
            The path to the folder where the dataset will be saved.
        """
        if self.batch_size is not None:
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
            temporary_folder = self._observer.temporary_folder
            # copy the content of the temporary folder to the folder where the dataset will be saved
            batches_directory = os.path.join(folder_path, "batches")
            if not os.path.exists(batches_directory):
                os.makedirs(os.path.join(batches_directory))

            for file in os.listdir(temporary_folder.name):
                if os.path.exists(batches_directory):
                    shutil.rmtree(batches_directory)
                shutil.copytree(os.path.join(temporary_folder.name, file), os.path.join(batches_directory, file))

            self.to_pickle(os.path.join(folder_path, f"{folder_path}.pkl"))

        else:
            self.to_pickle(os.path.join(f"{folder_path}.pkl"))

    @classmethod
    def load(cls, path: str):
        """
        Method to load the dataset.

        Parameters
        ----------
        path: str
            The path to the folder where the dataset is saved.
        """
        if os.path.isdir(path):
            dirs = os.listdir(path)

            if len(dirs) != 0:
                # create the dataset
                dataset = cls.from_pickle(os.path.join(path, f"{path}.pkl"))
                batches_directory = os.path.join(path, "batches")
                for folder in os.listdir(batches_directory):
                    if os.path.exists(os.path.join(dataset._observer.temporary_folder.name, folder)):
                        shutil.rmtree(os.path.join(dataset._observer.temporary_folder.name, folder))
                    shutil.copytree(os.path.join(batches_directory, folder),
                                    os.path.join(dataset._observer.temporary_folder.name, folder))
            else:
                raise ValueError("The directory is empty.")

        else:
            folder_path = path.replace(".pkl", "")
            dataset = cls.from_pickle(os.path.join(f"{folder_path}.pkl"))

        return dataset

    def __next__(self):
        """
        Method to iterate over the dataset.
        """
        if self.batch_size is not None and self._dataframe_generator is not None:
            self.notify(function="__next__")

            try:
                df = next(self._dataframe_generator)
            except StopIteration:
                self.end()
                self._dataframe_generator = None
                return False

            self.dataframe = df

            return True

        raise ValueError("The dataset is not iterable.")

    def next_batch(self):
        """
        Method to iterate over the dataset.
        """
        if self.batch_size is not None:
            self.notify(function="next_batch")
            if self._batch_state:
                return self
            else:
                self.end()
                return None
        else:
            raise ValueError("The dataset is not iterable.")

    @abstractmethod
    def select(self, ids: Union[List[str], List[int]], instance_type: str):
        """
        Method to select a subset of the dataset.

        Parameters
        ----------
        ids: List[str]
            The identifiers of the instances to select.
        instance_type: str
            The type of the instances to select.
        """

    @property
    @abstractmethod
    def dataframe(self):
        """
        Returns the dataset as a pandas dataframe.
        """

    @dataframe.setter
    @abstractmethod
    def dataframe(self, value):
        """
        Sets the dataset as a pandas dataframe.

        Parameters
        ----------
        value: pd.DataFrame
            The dataset as a pandas dataframe.
        """

    @property
    @abstractmethod
    def instances(self):
        """
        Returns the instances of the dataset.
        """

    @property
    @abstractmethod
    def identifiers(self):
        """
        Returns the identifiers of the instances of the dataset.
        """

    @property
    @abstractmethod
    def features(self) -> Dict[str, Dict[str, np.ndarray]]:
        """
        Returns the features of the dataset.
        """

    @features.setter
    @abstractmethod
    def features(self, value: Dict[str, Dict[str, np.ndarray]]):
        """
        Sets the features of the dataset.

        Parameters
        ----------
        value: Dict[str, Dict[str, np.ndarray]]
            The features of the dataset. With the instance type as key and a dictionary of features as value.
            The keys of the features dictionary are the instance identifiers and the values are the features.
            For single-instance datasets, the instance type is a placeholder.
        """

    @abstractmethod
    def add_features(self, instance_type: str, features: Dict[str, np.ndarray]):
        """
        Adds features to the dataset.

        Parameters
        ----------
        instance_type: str
            The instance type.
        features: Dict[str, np.ndarray]
            The features to add. With the instance identifiers as keys and the features as value.
        """

    @property
    @abstractmethod
    def labels(self):
        """
        Returns the labels of the dataset.
        """

    @property
    @abstractmethod
    def X(self) -> Dict[str, np.ndarray]:
        """
        Returns the features of the dataset.

        Returns
        -------
        Dict[str, np.ndarray]
            The features of the dataset. With the instance identifiers as keys and the features as value.
        """

    @property
    @abstractmethod
    def y(self) -> np.ndarray:
        """
        Returns the labels of the dataset.

        Returns
        -------
        np.ndarray
            The labels of the dataset.
        """

    @property
    @abstractmethod
    def features_fields(self):
        """
        Returns the features fields of the dataset.
        """

    @features_fields.setter
    @abstractmethod
    def features_fields(self, value: Any):
        """
        Sets the features fields of the dataset.

        Parameters
        ----------
        value: Any
            The features fields of the dataset.
        """

    @abstractmethod
    def get_instances(self, instance_type: str = None):
        """
        Returns the instances of the dataset.
        """

    def _set_dataframe(self, value: Any):
        """
        Private method to set the dataframe.
        Parameters
        ----------
        value: Any
            dataframe to be set, it can be in pd.DataFrame format, but can also be a List or Dictionary
            (it can be specific for each data type)
        """
        if isinstance(value, pd.DataFrame) or value is None:
            self._dataframe = value
        else:
            raise TypeError("It seems that the type of your input is not a pandas DataFrame."
                            "The type of the dataframe should be a pandas DataFrame")



from abc import abstractmethod

from pydantic import BaseModel
from pydantic.validators import dict_validator

from plants_sm.data_structures.dataset import Dataset, SingleInputDataset
from plants_sm.data_structures.dataset.single_input_dataset import PLACEHOLDER_FIELD
from plants_sm.estimation._utils import fit_status
from plants_sm.mixins.mixins import PickleMixin


class Estimator(BaseModel, PickleMixin):
    _fitted: bool = False

    class Config:
        """
        Model Configuration: https://pydantic-docs.helpmanual.io/usage/model_config/
        """
        extra = 'allow'
        allow_mutation = True
        validate_assignment = True
        underscore_attrs_are_private = True

    @classmethod
    def get_validators(cls):
        # yield dict_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        else:
            return cls(**dict_validator(value))

    @property
    def fitted(self):
        return self._fitted

    @fitted.setter
    def fitted(self, value: bool):
        if isinstance(value, bool):
            self._fitted = value

        else:
            raise TypeError("fitted has to be a boolean")

    @abstractmethod
    def _fit(self, dataset: Dataset, instance_type: str) -> 'Estimator':
        raise NotImplementedError

    @abstractmethod
    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'Estimator':
        raise NotImplementedError

    @fit_status
    def fit(self, dataset: Dataset, instance_type: str = None) -> 'Estimator':

        if dataset.batch_size is not None:
            batch = dataset.next_batch()
            while batch is not None:
                self._fit_dataset(batch, instance_type, batch=True)
                batch = dataset.next_batch()
        else:
            self._fit_dataset(dataset, instance_type)

        return self

    def _fit_dataset(self, dataset: Dataset, instance_type: str = None, batch=False) -> 'Estimator':
        """
        Fit the estimator on the dataset

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the estimator on
        instance_type: str
            type of the instances to fit. If None, it will only fit instances

        Returns
        -------
        fitted_estimator: Estimator
            fitted estimator
        """

        if batch:
            function = self._fit_batch
        else:
            function = self._fit

        if instance_type is None and isinstance(dataset, SingleInputDataset):
            function(dataset, PLACEHOLDER_FIELD)
        elif isinstance(dataset, SingleInputDataset):
            function(dataset, PLACEHOLDER_FIELD)
        else:
            function(dataset, instance_type)

        return self



from torch import nn


class CNN1D(nn.Module):
    def __init__(self, fc_hidden_sizes, cnn_layers, kernel_size, num_classes, last_sigmoid=True):
        super(CNN1D, self).__init__()

        self.fc_hidden_sizes = fc_hidden_sizes
        self.cnn_layers = cnn_layers

        self.last_sigmoid = last_sigmoid

        # Convolutional layer
        for i in range(0, len(cnn_layers)):
            if i == 0:
                setattr(self, f"conv{i}",
                        nn.Conv1d(in_channels=1, out_channels=cnn_layers[i], kernel_size=kernel_size[i]))
            else:
                setattr(self, f"conv{i}", nn.Conv1d(in_channels=cnn_layers[i - 1], out_channels=cnn_layers[i],
                                                    kernel_size=kernel_size[i]))
            setattr(self, f"pool{i}", nn.MaxPool1d(kernel_size=kernel_size[i]))

            setattr(self, f"relu_conv{i}", nn.ReLU())

        self.flatten = nn.Flatten()

        # Fully-connected input layer
        self.fc_initial = nn.LazyLinear(fc_hidden_sizes[0])
        self.relu_initial = nn.ReLU()

        # Fully-connected output layer
        for i in range(1, len(fc_hidden_sizes)):
            setattr(self, f"fc{i}", nn.Linear(fc_hidden_sizes[i - 1], fc_hidden_sizes[i]))
            setattr(self, f"relu{i}", nn.ReLU())

        self.fc_final = nn.Linear(fc_hidden_sizes[-1], num_classes)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        # Apply convolutional layer
        x = x[0]
        x = x.unsqueeze(1)
        for i in range(0, len(self.cnn_layers)):
            conv = getattr(self, f"conv{i}")
            x = conv(x)
            x = getattr(self, f"relu_conv{i}")(x)
            x = getattr(self, f"pool{i}")(x)

        x = self.flatten(x)

        x = self.fc_initial(x)
        x = self.relu_initial(x)
        # Apply fully-connected layer
        for i in range(1, len(self.fc_hidden_sizes)):
            x = getattr(self, f"fc{i}")(x)
            x = getattr(self, f"relu{i}")(x)

        x = self.fc_final(x)

        if self.last_sigmoid:
            x = self.sigmoid(x)

        return x



from copy import deepcopy
import torch
import torch.nn as nn

from plants_sm.models.pytorch_model import PyTorchModel


class DeepECCNN(nn.Module):
    def __init__(self, num_filters, input_size, num_columns, kernel_sizes, num_dense_layers, dense_layer_size,
                 num_classes):
        super(DeepECCNN, self).__init__()

        self.num_filters = num_filters
        self.kernel_sizes = kernel_sizes
        self.num_dense_layers = num_dense_layers

        for i in range(len(kernel_sizes)):
            setattr(self, f"conv{i}", nn.Conv2d(1, num_filters, kernel_size=(kernel_sizes[i], num_columns),
                                                stride=(1, 1), padding="valid"))
            setattr(self, f"maxpool{i}", nn.MaxPool2d((input_size - kernel_sizes[i], 1)))
            setattr(self, f"flatten{i}", nn.Flatten())
            setattr(self, f"batchnorm{i}", nn.BatchNorm1d(num_filters))

        self.concat = nn.Linear(num_filters * len(kernel_sizes), dense_layer_size)

        self.batchnorm4 = nn.BatchNorm1d(dense_layer_size)
        self.activation1 = nn.ReLU()

        self.last_sigmoid = True

        for i in range(num_dense_layers):
            setattr(self, f"dense{i}", nn.Linear(dense_layer_size, dense_layer_size))
            setattr(self, f"batchnorm_dense{i}", nn.BatchNorm1d(dense_layer_size))
            setattr(self, f"activation_dense{i}", nn.ReLU())

        self.dense_final = nn.Linear(dense_layer_size, num_classes)

    def forward(self, x):
        
        if type(x) == list:
            x = x[0]

        x = x.unsqueeze(1)
        to_concat = []
        for i in range(len(self.kernel_sizes)):
            x_copy = deepcopy(x)
            conv = getattr(self, f"conv{i}")
            x_copy = conv(x_copy)
            x_copy = getattr(self, f"maxpool{i}")(x_copy)
            x_copy = getattr(self, f"flatten{i}")(x_copy)
            x_copy = getattr(self, f"batchnorm{i}")(x_copy)
            to_concat.append(x_copy)

        x = torch.cat(to_concat, dim=1)
        x = self.concat(x)
        x = self.batchnorm4(x)
        x = self.activation1(x)

        for i in range(self.num_dense_layers):
            x = getattr(self, f"dense{i}")(x)
            x = getattr(self, f"batchnorm_dense{i}")(x)
            x = getattr(self, f"activation_dense{i}")(x)

        x = self.dense_final(x)
        if self.last_sigmoid:
            x = torch.sigmoid(x)

        return x


class DeepECCNNOptimal(DeepECCNN):

    def __init__(self, num_columns, input_size, num_classes):
        super().__init__(128, input_size, num_columns, [4, 8, 16], 2, 512, num_classes)


class DeepEC(PyTorchModel):

    def __init__(self, num_columns, input_size, num_classes,
                 loss_function, validation_loss_function,
                 batch_size,
                 optimizer=torch.optim.Adam, learning_rate=0.009999999776482582,
                 epochs=30, device="cuda:0", patience=4, **kwargs):
        self.num_columns = num_columns
        self.num_classes = num_classes
        self.input_size = input_size

        model = DeepECCNN(128, input_size, num_columns, [4, 8, 16], 2, 512, num_classes)
        self.optimizer = optimizer(params=model.parameters(), lr=learning_rate, betas=(0.9, 0.999), eps=1e-7)

        super().__init__(model=model,
                         loss_function=loss_function,
                         validation_loss_function=validation_loss_function,
                         optimizer=self.optimizer,
                         device=device,
                         epochs=epochs,
                         patience=patience,
                         batch_size=batch_size,
                         **kwargs
                         )



from plants_sm.featurization.proteins.bio_embeddings.esm_models import ESM1Model, ESM2Model
from plants_sm.parallelisation import TorchSpawner
from torch import nn
from tqdm import tqdm

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization._utils import call_set_features_names
from plants_sm.featurization.proteins.bio_embeddings.constants import ESM_DIMENSIONS, ESM_FUNCTIONS, ESM_LAYERS
from plants_sm.transformation.transformer import Transformer

from fairscale.nn.data_parallel import FullyShardedDataParallel as FSDP
from fairscale.nn.wrap import enable_wrap, wrap

import torch

from esm.pretrained import load_model_and_alphabet_local


class ESMEncoder(Transformer):
    """
    It encodes protein sequences with the embedding layer of the pre-trained model ESM-1B.
    The EsmEncoder operates only over pandas DataFrame.

    Parameters
    ----------
    batch_size: int, optional (default=16)
        The batch size to be used in the encoding process. Higher batch sizes can lead to OOM issues.
    esm_function: str, optional (default="esm1b_t33_650M_UR50S")
        The ESM function to be used.
    device: str, optional (default="cpu")
        The device to be used for the encoding process.
    num_gpus: int, optional (default=None)
        The number of GPUs to be used for the encoding process.
    output_dim: int, optional (default=2)
        The output dimension of the model.
    """

    batch_size: int = 16
    features_names = list = []
    esm_function: str = "esm2_t6_8M_UR50D"
    device: str = "cpu"
    num_gpus: int = None
    output_dim: int = 2
    return_contacts: bool = False
    model_path: str = None

    def set_features_names(self):
        """
        Set the features names of the encoded dataset.
        """
        self.features_names = [f"ESM_{self.esm_function}_{i}" for i in range(ESM_DIMENSIONS[self.esm_function])]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'ESMEncoder':
        """
        Fit the ESM. It loads the pre-trained model and the batch converter.

        Parameters
        ----------
        dataset: Dataset
            The dataset to be used to fit the Esm1bEncoder.
        instance_type: str
            The type of instance to be used to fit the ESM.

        Returns
        -------
        encoder: a fitted ESM
        """

        if self.esm_function in ESM_DIMENSIONS:

            esm_callable = ESM_FUNCTIONS[self.esm_function]
            self.layers = ESM_LAYERS[self.esm_function]

            if self.model_path is not None:
                model, self.alphabet = load_model_and_alphabet_local(self.model_path)
            else:
                model, self.alphabet = esm_callable()
            
            if self.num_gpus is not None and self.device == "cpu":
                self.device = "cuda"
                
            self.model = model
            self.batch_converter = self.alphabet.get_batch_converter()

            if self.num_gpus is not None and self.num_gpus > 1:
                self.is_ddf = True
            elif self.num_gpus is not None and self.num_gpus == 1:
                self.is_ddf = False
            else:
                self.num_gpus = 0
                self.is_ddf = False

            return self
        else:
            raise ValueError(f"Invalid esm_function. Available functions are: {list(ESM_DIMENSIONS.keys())}")

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'ESMEncoder':

        return self._fit(dataset, instance_type)

    @staticmethod
    def _generate_esm2_model(model: nn.Module,
                             layers: int,
                             instances: dict,
                             batch_size: int,
                             batch_converter: callable,
                             output_dim: int,
                             is_ddf: bool,
                             device: str = "cpu"):
        """
        Generate the ESM model.

        Parameters
        ----------
        model: nn.Module
            The ESM model.
        layers: int
            The number of layers of the ESM model.
        instances: dict
            The instances to be encoded.
        batch_size: int
            The batch size to be used in the encoding process.
        batch_converter: callable
            The batch converter to be used in the encoding process.
        output_dim: int
            The output dimension of the ESM model.
        is_ddf: bool
            Whether to use DDP or not.
        """

        if is_ddf:
            fsdp_params = dict(
                mixed_precision=True,
                flatten_parameters=True,
                state_dict_device=torch.device("cpu"),  # reduce GPU mem usage
                cpu_offload=False,  # enable cpu offloading
            )

            with enable_wrap(wrapper_cls=FSDP, **fsdp_params):
                model.eval()

                # Wrap each layer in FSDP separately
                for name, child in model.named_children():

                    if name == "layers":
                        for layer_name, layer in child.named_children():
                            wrapped_layer = wrap(layer)
                            setattr(child, layer_name, wrapped_layer)

                model = wrap(model)

        res = []
        batch = []
        batch_ids = []

        pbar = tqdm(desc="ESM", total=len(instances.items()))
        for instance_id, instance_representation in instances.items():

            batch.append((instance_id, instance_representation))
            batch_ids.append(instance_id)
            if len(batch) == batch_size:
                representations = {}
                _, _, batch_tokens = batch_converter(batch)
                if is_ddf:
                    batch_tokens = batch_tokens.cuda()
                else:
                    batch_tokens = batch_tokens.to(device)

                with torch.no_grad():
                    results = model(batch_tokens, repr_layers=[layers], return_contacts=False)
                    results["representations"][layers] = results["representations"][layers].cpu().detach().numpy()
                    representations['representations'] = results["representations"][layers]

                    for i, batch_instance_id in enumerate(batch_ids):

                        if output_dim == 2:
                            res.append((batch_instance_id,
                                        representations['representations'][i, 1: len(batch[i][1]) + 1].mean(0)))
                        else:
                            res.append((batch_instance_id,
                                        representations['representations'][i, 1: len(batch[i][1]) + 1]))
                    if is_ddf:
                        torch.cuda.empty_cache()
                    batch = []
                    batch_ids = []
                    pbar.update(batch_size)

        if len(batch) != 0:

            representations = {}
            _, _, batch_tokens = batch_converter(batch)

            if is_ddf:
                batch_tokens = batch_tokens.cuda()
            else:
                batch_tokens = batch_tokens.to(device)

            results = model(batch_tokens, repr_layers=[layers], return_contacts=False)
            results["representations"][layers] = results["representations"][layers].cpu().detach().numpy()
            representations['representations'] = results["representations"][layers]

            for i, batch_instance_id in enumerate(batch_ids):
                if output_dim == 2:
                    res.append((batch_instance_id,
                                representations['representations'][i, 1: len(batch[i][1]) + 1].mean(0)))
                else:
                    res.append((batch_instance_id,
                                representations['representations'][i, 1: len(batch[i][1]) + 1]))

            torch.cuda.empty_cache()
            pbar.update(len(batch_ids))

        return res

    @call_set_features_names
    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        It encodes a protein sequence with the embedding layer of the pre-trained model ESM-1B.

        Parameters
        ----------
        dataset: Dataset
            The dataset to be used to encode the protein sequences.
        instance_type: str
            The instance type to be encoded.

        Returns
        -------
        encoded_sequence: np.ndarray
            The encoded protein sequence.
        """

        instances = dataset.get_instances(instance_type)

        # initialize the model with FSDP wrapper
        if "esm2" in self.esm_function:
            if self.num_gpus == 1 and self.device != "cpu":
                devices = [self.device]
            else:
                devices = None

            model = ESM2Model(alphabet=self.alphabet, num_layers=self.model.num_layers, embed_dim=self.model.embed_dim,
                              attention_heads=self.model.attention_heads, token_dropout=self.model.token_dropout,
                              is_ddf=self.is_ddf, num_gpus=self.num_gpus, devices=devices)
            model.load_state_dict(self.model.state_dict())

        else:
            model = ESM1Model(self.model.args, alphabet=self.alphabet,
                               device=self.device)
            model.load_state_dict(self.model.state_dict())

        if self.is_ddf:
            res = TorchSpawner().run(self._generate_esm2_model,
                                     model=model,
                                     layers=self.layers,
                                     instances=instances,
                                     batch_size=self.batch_size,
                                     batch_converter=self.batch_converter,
                                     output_dim=self.output_dim,
                                     is_ddf=self.is_ddf,
                                     device=self.device)

        else:
            res = self._generate_esm2_model(model,
                                            layers=self.layers,
                                            instances=instances,
                                            batch_size=self.batch_size,
                                            batch_converter=self.batch_converter,
                                            output_dim=self.output_dim,
                                            is_ddf=self.is_ddf,
                                            device=self.device)

        dataset.add_features(instance_type, dict(res))

        dataset.features_fields[instance_type] = self.features_names

        return dataset



import torch
from torch import nn
import torch.nn.functional as F

from plants_sm.models.pytorch_model import PyTorchModel


class DSPACEModel(nn.Module):

    def __init__(self, num_aa, sequence_dimension, num_classes):
        super().__init__()
        self.conv0 = nn.Conv1d(in_channels=num_aa, out_channels=16, kernel_size=3, padding="same")
        self.bn0 = nn.BatchNorm1d(16)
        self.conv1 = nn.Conv1d(in_channels=16, out_channels=24, kernel_size=3, padding="same")
        self.bn1 = nn.BatchNorm1d(24)
        self.maxpool1 = nn.MaxPool1d(kernel_size=2)

        self.conv2 = nn.Conv1d(in_channels=24, out_channels=32, kernel_size=5, padding="same")
        self.bn2 = nn.BatchNorm1d(32)
        self.conv3 = nn.Conv1d(in_channels=32, out_channels=48, kernel_size=5, padding="same")
        self.bn3 = nn.BatchNorm1d(48)
        self.maxpool2 = nn.MaxPool1d(kernel_size=2)

        self.conv4 = nn.Conv1d(in_channels=48, out_channels=64, kernel_size=7, padding="same")
        self.bn4 = nn.BatchNorm1d(64)
        self.conv5 = nn.Conv1d(in_channels=64, out_channels=96, kernel_size=7, padding="same")
        self.bn5 = nn.BatchNorm1d(96)
        self.maxpool3 = nn.MaxPool1d(kernel_size=2)

        self.fc0 = nn.Linear(96 * (sequence_dimension // 8), 2048)
        self.bn6 = nn.BatchNorm1d(2048)
        self.fc1 = nn.Linear(2048, 1024)
        self.bn7 = nn.BatchNorm1d(1024)
        self.fc2 = nn.Linear(1024, 512)
        self.bn8 = nn.BatchNorm1d(512)
        self.fc3 = nn.Linear(512, 256)
        self.embedding_bn = nn.BatchNorm1d(256)

        self.classification_layer = nn.Linear(256, num_classes)

    def forward(self, x):
        if type(x) == list:
            x = x[0]
        
        x = x.permute(0, 2, 1)
        x = F.elu(self.bn0(self.conv0(x)))
        x = F.elu(self.bn1(self.conv1(x)))
        x = self.maxpool1(x)

        x = F.elu(self.bn2(self.conv2(x)))
        x = F.elu(self.bn3(self.conv3(x)))
        x = self.maxpool2(x)

        x = F.elu(self.bn4(self.conv4(x)))
        x = F.elu(self.bn5(self.conv5(x)))
        x = self.maxpool3(x)

        x = x.view(x.size(0), -1)  # Flatten
        x = F.elu(self.bn6(self.fc0(x)))
        x = F.elu(self.bn7(self.fc1(x)))
        x = F.elu(self.bn8(self.fc2(x)))
        x = F.elu(self.fc3(x))
        x = self.embedding_bn(x)

        x = torch.sigmoid(self.classification_layer(x))

        return x


class DSPACE(PyTorchModel):

    def __init__(self, num_columns, input_size, num_classes,
                 loss_function, validation_loss_function,
                 batch_size,
                 optimizer=torch.optim.NAdam, learning_rate=0.001,
                 epochs=30, device="cuda:0", patience=4, **kwargs):
        model = DSPACEModel(num_columns, input_size, num_classes)
        self.optimizer = optimizer(params=model.parameters(), lr=learning_rate)

        super().__init__(model=model,
                         loss_function=loss_function,
                         validation_loss_function=validation_loss_function,
                         optimizer=self.optimizer,
                         device=device,
                         epochs=epochs,
                         patience=patience,
                         batch_size=batch_size,
                         **kwargs
                         )



from abc import abstractmethod
from typing import Any

import optuna

from plants_sm.data_structures.dataset import Dataset


class Experiment:

    def __init__(self, train_dataset: Dataset, validation_dataset: Dataset = None, **kwargs):
        """
        Initialize an experiment.

        Parameters
        ----------
        kwargs: dict
            Keyword arguments to pass to optuna.create_study.
            storage:
                Database URL. If this argument is set to None, in-memory storage is used, and the
                :class:`~optuna.study.Study` will not be persistent.

                .. note::
                    When a database URL is passed, Optuna internally uses `SQLAlchemy`_ to handle
                    the database. Please refer to `SQLAlchemy's document`_ for further details.
                    If you want to specify non-default options to `SQLAlchemy Engine`_, you can
                    instantiate :class:`~optuna.storages.RDBStorage` with your desired options and
                    pass it to the ``storage`` argument instead of a URL.

                 .. _SQLAlchemy: https://www.sqlalchemy.org/
                 .. _SQLAlchemy's document:
                     https://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls
                 .. _SQLAlchemy Engine: https://docs.sqlalchemy.org/en/latest/core/engines.html
            sampler:
                A sampler object that implements background algorithm for value suggestion.
                If :obj:`None` is specified, :class:`~optuna.samplers.TPESampler` is used during
                single-objective optimization and :class:`~optuna.samplers.NSGAIISampler` during
                multi-objective optimization. See also :class:`~optuna.samplers`.
            pruner:
                A pruner object that decides early stopping of unpromising trials. If :obj:`None`
                is specified, :class:`~optuna.pruners.MedianPruner` is used as the default. See
                also :class:`~optuna.pruners`.
            study_name:
                Study's name. If this argument is set to None, a unique name is generated
                automatically.
            direction:
                Direction of optimization. Set ``minimize`` for minimization and ``maximize`` for
                maximization. You can also pass the corresponding :class:`~optuna.study.StudyDirection`
                object. ``direction`` and ``directions`` must not be specified at the same time.

                .. note::
                    If none of `direction` and `directions` are specified, the direction of the study
                    is set to "minimize".
            load_if_exists:
                Flag to control the behavior to handle a conflict of study names.
                In the case where a study named ``study_name`` already exists in the ``storage``,
                a :class:`~optuna.exceptions.DuplicatedStudyError` is raised if ``load_if_exists`` is
                set to :obj:`False`.
                Otherwise, the creation of the study is skipped, and the existing one is returned.
            directions:
                A sequence of directions during multi-objective optimization.
                ``direction`` and ``directions`` must not be specified at the same time.
        """
        self.best_parameters = None
        self.train_dataset = train_dataset
        self.validation_dataset = validation_dataset
        self.study = optuna.create_study(**kwargs)

    @property
    @abstractmethod
    def best_experiment(self) -> Any:
        """
        Method to be implemented by all experiments to return the best experiment. It can be a Pipeline, a Model,
        a Transformer, etc.

        Returns
        -------
        Any
            The best experiment. It can be a Pipeline, a Model, Transformer, etc.
        """

    @abstractmethod
    def objective(self, trial: optuna.trial.Trial) -> float:
        """
        Method to be implemented by all experiments to define the objective function.

        Parameters
        ----------
        trial: optuna.trial.Trial
            A trial object that contains the current suggested hyperparameters.

        Returns
        -------
        float
            The value of the objective function.
        """

    def run(self, **kwargs):
        """
        Run the experiment.

        Parameters
        ----------
        kwargs: dict
            Keyword arguments to pass to optuna.Study.optimize.
            n_trials:
                The number of trials for each process. :obj:`None` represents no limit in terms of
                the number of trials. The study continues to create trials until the number of
                trials reaches ``n_trials``, ``timeout`` period elapses,
                :func:`~optuna.study.Study.stop` is called, or a termination signal such as
                SIGTERM or Ctrl+C is received.

                .. seealso::
                    :class:`optuna.study.MaxTrialsCallback` can ensure how many times trials
                    will be performed across all processes.
            timeout:
                Stop study after the given number of second(s). :obj:`None` represents no limit in
                terms of elapsed time. The study continues to create trials until the number of
                trials reaches ``n_trials``, ``timeout`` period elapses,
                :func:`~optuna.study.Study.stop` is called or, a termination signal such as
                SIGTERM or Ctrl+C is received.
            n_jobs:
                The number of parallel jobs. If this argument is set to ``-1``, the number is
                set to CPU count.

                .. note::
                    ``n_jobs`` allows parallelization using :obj:`threading` and may suffer from
                    `Python's GIL <https://wiki.python.org/moin/GlobalInterpreterLock>`_.
                    It is recommended to use :ref:`process-based parallelization<distributed>`
                    if ``func`` is CPU bound.

            catch:
                A study continues to run even when a trial raises one of the exceptions specified
                in this argument. Default is an empty tuple, i.e. the study will stop for any
                exception except for :class:`~optuna.exceptions.TrialPruned`.
            callbacks:
                List of callback functions that are invoked at the end of each trial. Each function
                must accept two parameters with the following types in this order:
                :class:`~optuna.study.Study` and :class:`~optuna.trial.FrozenTrial`.

                .. seealso::

                    See the tutorial of :ref:`optuna_callback` for how to use and implement
                    callback functions.

            gc_after_trial:
                Flag to determine whether to automatically run garbage collection after each trial.
                Set to :obj:`True` to run the garbage collection, :obj:`False` otherwise.
                When it runs, it runs a full collection by internally calling :func:`gc.collect`.
                If you see an increase in memory consumption over several trials, try setting this
                flag to :obj:`True`.

                .. seealso::

                    :ref:`out-of-memory-gc-collect`

            show_progress_bar:
                Flag to show progress bars or not. To disable progress bar, set this :obj:`False`.
                Currently, progress bar is experimental feature and disabled
                when ``n_trials`` is :obj:`None`, ``timeout`` not is :obj:`None`, and
                ``n_jobs`` :math:`\\ne 1`.

        """
        self.study.optimize(self.objective, **kwargs)
        self.best_parameters = self.study.best_params



from abc import abstractmethod

from plants_sm.data_structures.dataset import Dataset, SingleInputDataset
from plants_sm.data_structures.dataset.single_input_dataset import PLACEHOLDER_FIELD
from plants_sm.estimation.estimator import Estimator


class Transformer(Estimator):
    n_jobs: int = 1

    @abstractmethod
    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        Abstract method that has to be implemented by all transformers
        """
        raise NotImplementedError

    def transform(self, dataset: Dataset, instance_type: str = None) -> Dataset:
        """
        Transform the dataset according to the transformer

        Parameters
        ----------
        dataset: Dataset
            dataset to transform
        instance_type: str
            type of the instances to transform. If None, it will only transform instances
            if the dataset has a single input

        Returns
        -------
        transformed_dataset: Dataset
            transformed dataset
        """
        if self.fitted:
            if dataset.batch_size is not None:
                dataset.reset_batch()
                batch = dataset.next_batch()
                while batch is not None:
                    self._transform_dataset(dataset, instance_type)
                    batch = dataset.next_batch()
            else:
                return self._transform_dataset(dataset, instance_type)

        else:
            # TODO : implement exception
            raise Exception("The transformer has to be fitted before transforming the dataset")

    def _transform_dataset(self, dataset: Dataset, instance_type: str = None) -> Dataset:
        """
        Transform the dataset according to the transformer

        Parameters
        ----------
        dataset: Dataset
            dataset to transform
        instance_type: str
            type of the instances to transform. If None, it will only transform instances

        Returns
        -------
        transformed_dataset: Dataset
            transformed dataset
        """
        if instance_type is None and isinstance(dataset, SingleInputDataset):
            return self._transform(dataset, PLACEHOLDER_FIELD)
        elif isinstance(dataset, SingleInputDataset):
            return self._transform(dataset, PLACEHOLDER_FIELD)
        else:
            return self._transform(dataset, instance_type)

    def fit_transform(self, dataset: Dataset, instance_type: str = None) -> Dataset:
        """
        Fit the transformer and transform the dataset
        """

        return self.fit(dataset, instance_type).transform(dataset, instance_type)



import os
from typing import Union

from joblib import Parallel, delayed
from tqdm import tqdm
from plants_sm.featurization.proteins.bio_embeddings.constants import ESM_FUNCTIONS, ESM_LAYERS
from plants_sm.models.fc.fc import DNN
import lightning as L
from plants_sm.models.lightning_model import InternalLightningModule

from plants_sm.models.ec_number_prediction._esm_model_architectures import FineTuneESM2

from torch.utils.data import DataLoader, TensorDataset

from plants_sm.data_structures.dataset import Dataset, PLACEHOLDER_FIELD
from plants_sm.featurization.proteins.bio_embeddings.constants import ESM_FUNCTIONS

import torch

from torch import nn


from torch.utils.data import TensorDataset, DataLoader
import lightning as L
import torch

from plants_sm.transformation._utils import tqdm_joblib 
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.lr_scheduler import LambdaLR

class EC_ESM_Lightning(InternalLightningModule):

    def __init__(self, model_name, hidden_layers, num_classes, freeze_layers=True, use_embedding = False, **kwargs):
        super(EC_ESM_Lightning, self).__init__(**kwargs)
        self.model_name = model_name
        
        self._contructor_parameters.update({"model_name": model_name, 
                                      "hidden_layers": hidden_layers, 
                                      "num_classes": num_classes})
        esm_callable = ESM_FUNCTIONS[model_name]
        self.layers = ESM_LAYERS[model_name]

        self.use_embedding = use_embedding
        self.freeze_layers = freeze_layers

        model, self.alphabet = esm_callable()

        self.batch_converter = self.alphabet.get_batch_converter()
        self.esm_model = FineTuneESM2(alphabet=self.alphabet, num_layers=model.num_layers, embed_dim=model.embed_dim,
                                      attention_heads=model.attention_heads, token_dropout=model.token_dropout)
        
        self.esm_model.load_state_dict(model.state_dict())

        self.dnn = DNN(model.embed_dim, hidden_layers, num_classes, batch_norm=True, last_sigmoid=True)


    def training_step(self, batch, batch_idx):
        x, y = batch

        if self.freeze_layers:
            if self.trainer.current_epoch < 14:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers)
            elif self.trainer.current_epoch >= 14 and self.trainer.current_epoch < 29:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers*3//4)
            else:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers*2//3)
        else:
            output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True)

        output = output["representations"][self.layers]
        if not self.use_embedding:
            output = output[:, 0, :]

        else:
            batch_lens = (x != self.alphabet.padding_idx).sum(1)
            sequence_representations = []
            for i, tokens_len in enumerate(batch_lens):
                sequence_representations.append(output[i, 1 : tokens_len - 1].mean(0))

            output = torch.stack(sequence_representations)
        logits = self.dnn([output])

        loss = self.compute_loss(logits, y)
        
        self.training_step_outputs.append(logits)
        self.training_step_y_true.append(y)
        self.log("train_loss", loss.item(), on_epoch=True, 
                 prog_bar=True, logger=True, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target = batch

        output = self.esm_model(inputs, repr_layers=[self.layers], need_head_weights=True)
        output = output["representations"][self.layers]

        if not self.use_embedding:
            output = output[:, 0, :]

        else:
            batch_lens = (inputs != self.alphabet.padding_idx).sum(1)
            sequence_representations = []
            for i, tokens_len in enumerate(batch_lens):
                sequence_representations.append(output[i, 1 : tokens_len - 1].mean(0))

            output = torch.stack(sequence_representations)

        output = self.dnn([output])
        self.validation_step_outputs.append(output)
        self.validation_step_y_true.append(target)

    # def forward(self, data):

    #     output = self.esm_model(data, repr_layers=[self.layers], need_head_weights=True)
    #     output = output["representations"][self.layers]
    #     output = output[:, 0, :]
    #     x = self.dnn([output])

    #     return x


    def configure_optimizers(self):
        optimizer = torch.optim.Adam([{'params': self.esm_model.parameters(), 'lr': 1e-7},
                                      {'params': self.dnn.parameters()}], lr=1e-3)
        scheduler = ReduceLROnPlateau(optimizer, 'min')

        # Define a custom learning rate scheduler using LambdaLR
        scheduler = {'scheduler': ReduceLROnPlateau(optimizer, 'min'), 'monitor': 'val_loss'}

        return [optimizer], [scheduler]

    
    def compute_loss(self, logits, targets):
        return nn.BCELoss()(logits, targets)
    
    def contructor_parameters(self):
        return self._contructor_parameters
    
class EC_ESM1b_Lightning(InternalLightningModule):
    def __init__(self, hidden_layers, num_classes, freeze_layers,use_embedding, **kwargs):
        super(EC_ESM1b_Lightning, self).__init__(**kwargs)
        esm_callable = ESM_FUNCTIONS["esm1b_t33_650M_UR50S"]
        self.layers = ESM_LAYERS["esm1b_t33_650M_UR50S"]

        self.model_name = "esm1b_t33_650M_UR50S"

        self.use_embedding = use_embedding
        self.freeze_layers = freeze_layers

        model, self.alphabet = esm_callable()

        self.batch_converter = self.alphabet.get_batch_converter()
        self.esm_model = model

        self.dnn = DNN(1280, hidden_layers, num_classes, batch_norm=False, last_sigmoid=True)

    def training_step(self, batch, batch_idx):
        x, y = batch

        if self.freeze_layers:
            if self.trainer.current_epoch < 14:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers)
            elif self.trainer.current_epoch >= 14 and self.trainer.current_epoch < 29:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers*3//4)
            else:
                output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True, layers_to_freeze=self.layers*2//3)
        else:
            output = self.esm_model(x, repr_layers=[self.layers], need_head_weights=True)

        output = output["representations"][self.layers]
        if not self.use_embedding:
            output = output[:, 0, :]

        else:
            batch_lens = (x != self.alphabet.padding_idx).sum(1)
            sequence_representations = []
            for i, tokens_len in enumerate(batch_lens):
                sequence_representations.append(output[i, 1 : tokens_len - 1].mean(0))

            output = torch.stack(sequence_representations)
        logits = self.dnn([output])

        loss = self.compute_loss(logits, y)
        
        self.training_step_outputs.append(logits)
        self.training_step_y_true.append(y)
        self.log("train_loss", loss.item(), on_epoch=True, 
                 prog_bar=True, logger=True, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target = batch

        output = self.esm_model(inputs, repr_layers=[self.layers], need_head_weights=True)
        output = output["representations"][self.layers]

        if not self.use_embedding:
            output = output[:, 0, :]

        else:
            batch_lens = (inputs != self.alphabet.padding_idx).sum(1)
            sequence_representations = []
            for i, tokens_len in enumerate(batch_lens):
                sequence_representations.append(output[i, 1 : tokens_len - 1].mean(0))

            output = torch.stack(sequence_representations)

        output = self.dnn([output])
        self.validation_step_outputs.append(output)
        self.validation_step_y_true.append(target)
    
    def configure_optimizers(self):
        optimizer = torch.optim.Adam([{'params': self.esm_model.parameters(), 'lr': 1e-5},
                                      {'params': self.dnn.parameters()}], lr=1e-3)
        scheduler = ReduceLROnPlateau(optimizer, 'min')

        # Define a custom learning rate scheduler using LambdaLR
        scheduler = {'scheduler': ReduceLROnPlateau(optimizer, 'min'), 'monitor': 'val_loss'}

        return [optimizer], [scheduler]
    
    def compute_loss(self, logits, targets):
        return nn.BCELoss()(logits, targets)

    def _preprocess_data(self, dataset: Dataset, shuffle: bool = True) -> DataLoader:
        tensors = []
        sequences = [(sequence_id, dataset.instances[PLACEHOLDER_FIELD][sequence_id]) for sequence_id in
                     dataset.dataframe[dataset.instances_ids_field]]
        
        esm_callable = ESM_FUNCTIONS[self.model_name]

        _, alphabet = esm_callable()
        batch_converter = alphabet.get_batch_converter()

        # _, _, tokens = batch_converter(sequences)

        batch_size = 10000  # You can adjust this based on your preferences

        # Initialize the progress bar
        progress_bar = tqdm(total=len(sequences), desc="Processing sequences", position=0, leave=True)

        # Define the function to be parallelized
        def process_batch(batch):
            _, _, tokens = batch_converter(batch)
            return tokens

        # Process sequences in parallel with a progress bar in batches
        result_tokens = []
        for i in range(0, len(sequences), batch_size):
            batch = sequences[i:i + batch_size]
            batch_results = Parallel(n_jobs=-1)(delayed(process_batch)(batch) for batch in [batch])
            result_tokens.extend(batch_results)
            progress_bar.update(len(batch))

        # Close the progress bar
        progress_bar.close()

        # Use joblib to parallelize the function across sequences
        result_tokens = torch.cat(result_tokens, dim=0)
        tensors.append(result_tokens)

        try:
            if dataset.y is not None:
                tensors.append(torch.tensor(dataset.y, dtype=torch.float))
        except ValueError:
            pass

        dataset = TensorDataset(
            *tensors
        )
        data_loader = DataLoader(
            dataset,
            shuffle=shuffle,
            batch_size=self.batch_size
        )
        return data_loader
    


import numpy as np
from map4 import MAP4Calculator
from rdkit.Chem import MolFromSmiles

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator


class MAP4Fingerprint(FeaturesGenerator):
    dimensions = 1024
    radius = 2
    is_counted = False

    def set_features_names(self):
        self.features_names = [f"map4_fingerprint_{i}" for i in range(self.dimensions)]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'FeaturesGenerator':
        self.fingerprint = MAP4Calculator(self.dimensions, self.radius, self.is_counted, True)
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'FeaturesGenerator':
        return self._fit(dataset, instance_type)

    def _featurize(self, molecule: str) -> np.ndarray:
        mol = MolFromSmiles(molecule)
        map4_fingerprint = self.fingerprint.calculate(mol)
        map4_fingerprint = map4_fingerprint.astype("int8")
        try:
            assert map4_fingerprint.shape == (self.dimensions,)
        except AssertionError:
            map4_fingerprint = np.zeros(self.dimensions, dtype=np.int8)
        return map4_fingerprint



import os
import pickle
import torch.distributed as dist
import torch.multiprocessing as mp
from datetime import timedelta

import torch

import tempfile

DEFAULT_TIMEOUT = timedelta(seconds=10)


class TorchSpawner:

    def __init__(self, backend: str = "NCCL", num_machines: int = 1, machine_rank: int = 0,
                 dist_url: str = "tcp://127.0.0.1:1234") -> None:
        """
        TorchSpawner is a class that allows to spawn a torch process in a distributed way.

        Parameters
        ----------
        backend: str
            The backend to be used in the distributed process.
        num_machines: int
            The number of machines to be used in the distributed process.
        machine_rank: int
            The rank of the machine to be used in the distributed process.
        dist_url: str
            The url to be used in the distributed process.
        """
        self.backend = backend
        # self.num_gpus = num_gpus
        self.num_machines = num_machines
        self.machine_rank = machine_rank
        self.dist_url = dist_url

    def run(self, main_func, **kwargs):
        """
        Run the distributed process.

        Parameters
        ----------
        main_func: callable
            The function to be run in the distributed process.
        kwargs: dict
            The arguments to be passed to the main function.
        """
        torch.set_num_threads(1)
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["NCCL_DEBUG"] = "ERROR"

        world_size = 1

        results_file = tempfile.NamedTemporaryFile(delete=True)

        mp.spawn(
            TorchSpawner.distributed_worker,
            nprocs=1,
            args=(
                main_func,
                self.backend,
                world_size,
                1,
                self.machine_rank,
                self.dist_url,
                results_file.name,
                kwargs
            ),
            daemon=False,
        )

        results = pickle.load(results_file)
        results_file.close()
        return results

    @staticmethod
    def distributed_worker(
            local_rank,
            main_func,
            backend,
            world_size,
            num_gpus_per_machine,
            machine_rank,
            dist_url,
            results_file,
            kwargs
    ):
        """
        Run the distributed process.
        """
        LOCAL_PROCESS_GROUP = None

        if not torch.cuda.is_available():
            raise RuntimeError("CUDA is not available. Please check your installation.")

        global_rank = machine_rank * num_gpus_per_machine + local_rank
        try:
            dist.init_process_group(
                backend=backend,
                init_method=dist_url,
                world_size=world_size,
                rank=global_rank,
                timeout=DEFAULT_TIMEOUT,
            )
        except Exception as e:
            raise e

        dist.barrier()

        if num_gpus_per_machine > torch.cuda.device_count():
            raise RuntimeError
        torch.cuda.set_device(local_rank)

        # Setup the local process group (which contains ranks within the same machine)
        if LOCAL_PROCESS_GROUP is not None:
            raise RuntimeError

        num_machines = world_size // num_gpus_per_machine

        for idx in range(num_machines):
            ranks_on_i = list(range(idx * num_gpus_per_machine, (idx + 1) * num_gpus_per_machine))
            pg = dist.new_group(ranks_on_i)
            if idx == machine_rank:
                LOCAL_PROCESS_GROUP = pg

        results = main_func(**kwargs)
        with open(results_file, "wb") as results_file:
            pickle.dump(results, results_file)



import esm

from typing import Union
import torch
from esm.model.esm2 import ESM2
from esm.model.esm1 import ProteinBertModel


class ESM2Model(ESM2):

    def __init__(self, num_layers: int = 33,
                 embed_dim: int = 1280,
                 attention_heads: int = 20,
                 alphabet: Union[esm.data.Alphabet, str] = "ESM-1b",
                 token_dropout: bool = True, is_ddf=False,
                 num_gpus=1, 
                 devices=None) -> None:
        self.is_ddf = is_ddf
        self.num_gpus = num_gpus
        if devices is None and num_gpus is not None:
            self.devices = [f"cuda:{i}" for i in range(num_gpus)]
        else:
            self.devices = devices
        super().__init__(num_layers, embed_dim, attention_heads, alphabet, token_dropout)

    def forward(self, tokens, repr_layers=None, need_head_weights=False, return_contacts=False):

        if repr_layers is None:
            repr_layers = []
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 2
        padding_mask = tokens.eq(self.padding_idx)  # B, T

        self.embed_tokens = self.embed_tokens.to(tokens.device)
        x = self.embed_scale * self.embed_tokens(tokens)

        if self.token_dropout:
            x.masked_fill_((tokens == self.mask_idx).unsqueeze(-1), 0.0)
            # x: B x T x C
            mask_ratio_train = 0.15 * 0.8
            src_lengths = (~padding_mask).sum(-1)
            mask_ratio_observed = (tokens == self.mask_idx).sum(-1).to(x.dtype) / src_lengths
            x = x * (1 - mask_ratio_train) / (1 - mask_ratio_observed)[:, None, None]

        if padding_mask is not None:
            x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            attn_weights = []

        # (B, T, E) => (T, B, E)
        x = x.transpose(0, 1)

        if not padding_mask.any():
            padding_mask = None

        if self.is_ddf:
            gpus = self.devices
        i = 0
        for layer_idx, layer in enumerate(self.layers):
            if self.is_ddf:
                gpu = gpus[i % len(gpus)]
                x.to(gpu, dtype=torch.float32)
                layer.to(gpu)
            elif self.num_gpus == 0:
                x = x.to("cpu")
                layer = layer.to("cpu")
            else:
                x = x.to(self.devices[0])
                layer = layer.to(self.devices[0])

            x, attn = layer(
                x,
                self_attn_padding_mask=padding_mask,
                need_head_weights=need_head_weights,
            )

            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.transpose(0, 1)
            if need_head_weights:
                # (H, B, T, T) => (B, H, T, T)
                attn_weights.append(attn.transpose(1, 0))
            i += 1

        if self.is_ddf:
            gpu = gpus[i % len(gpus)]
            x.to(gpu)
            self.emb_layer_norm_after.to(gpu)
        elif self.num_gpus == 0:
            x = x.to("cpu")
            self.emb_layer_norm_after = self.emb_layer_norm_after.to("cpu")
        else:
            x = x.to(self.devices[0])
            self.emb_layer_norm_after = self.emb_layer_norm_after.to(self.devices[0])

        i += 1
        x = self.emb_layer_norm_after(x)
        x = x.transpose(0, 1)  # (T, B, E) => (B, T, E)

        # last hidden representation should have layer norm applied
        if (layer_idx + 1) in repr_layers:
            hidden_representations[layer_idx + 1] = x

        if self.is_ddf:
            gpu = gpus[i % len(gpus)]
            x.to(gpu)
            self.lm_head.to(gpu)
        elif self.num_gpus == 0:
            x = x.to("cpu")
            self.lm_head = self.lm_head.to("cpu")
        else:
            x = x.to(self.devices[0])
            self.lm_head = self.lm_head.to(self.devices[0])
        i += 1
        x = self.lm_head(x)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # attentions: B x L x H x T x T
            attentions = torch.stack(attn_weights, 1)
            if padding_mask is not None:
                attention_mask = 1 - padding_mask.type_as(attentions)
                attention_mask = attention_mask.unsqueeze(1) * attention_mask.unsqueeze(2)
                attentions = attentions * attention_mask[:, None, None, :, :]
            result["attentions"] = attentions
            if return_contacts:
                contacts = self.contact_head(tokens, attentions)
                result["contacts"] = contacts

        return result


# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import torch
import torch.nn.functional as F


class ESM1Model(ProteinBertModel):

    def __init__(self, args, alphabet, device="cpu"):
        super().__init__(args, alphabet)
        self.device = device
        for param in self.parameters():
            # Check if parameter dtype is  Half (float16)
            if param.dtype == torch.float16:
                param.data = param.data.to(torch.float32)

    def forward(self, tokens, repr_layers=[], need_head_weights=False, return_contacts=False):
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 2
        padding_mask = tokens.eq(self.padding_idx)  # B, T

        self.embed_tokens = self.embed_tokens.to(tokens.device)

        x = self.embed_scale * self.embed_tokens(tokens)
        x = x.float()

        if getattr(self.args, "token_dropout", False):
            x.masked_fill_((tokens == self.mask_idx).unsqueeze(-1), 0.0)
            # x: B x T x C
            mask_ratio_train = 0.15 * 0.8
            src_lengths = (~padding_mask).sum(-1)
            mask_ratio_observed = (tokens == self.mask_idx).sum(-1).float() / src_lengths
            x = x * (1 - mask_ratio_train) / (1 - mask_ratio_observed)[:, None, None]

        self.embed_positions = self.embed_positions.to(tokens.device)
        x = x + self.embed_positions(tokens)


        if self.model_version == "ESM-1b":
            if self.emb_layer_norm_before:
                x = x.to(self.device)
                self.emb_layer_norm_before = self.emb_layer_norm_before.to(self.device)
                self.emb_layer_norm_before.weight = self.emb_layer_norm_before.weight.float()
                self.emb_layer_norm_before.bias = self.emb_layer_norm_before.bias.float()
                x = self.emb_layer_norm_before(x)
            if padding_mask is not None:
                x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            attn_weights = []

        # (B, T, E) => (T, B, E)
        x = x.transpose(0, 1)

        if not padding_mask.any():
            padding_mask = None

        for layer_idx, layer in enumerate(self.layers):

            x = x.to(self.device)
            layer.to(self.device)
            if padding_mask is not None:
                padding_mask = padding_mask.to(self.device)

            layer.buffer_dtype = torch.float32
            layer.compute_dtype = torch.float32
            x, attn = layer(
                x, self_attn_padding_mask=padding_mask, need_head_weights=need_head_weights
            )
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.transpose(0, 1)
            if need_head_weights:
                # (H, B, T, T) => (B, H, T, T)
                attn_weights.append(attn.transpose(1, 0))

        if self.model_version == "ESM-1b":
            x = x.to(self.device)
            self.emb_layer_norm_after = self.emb_layer_norm_after.to(self.device)

            self.emb_layer_norm_after.weight = self.emb_layer_norm_after.weight.float()
            self.emb_layer_norm_after.bias = self.emb_layer_norm_after.bias.float()
            x = self.emb_layer_norm_after(x)
            x = x.transpose(0, 1)  # (T, B, E) => (B, T, E)

            # last hidden representation should have layer norm applied
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x

            x = x.to(self.device)
            self.lm_head = self.lm_head.to(self.device)

            self.lm_head.weight = self.lm_head.weight.float()
            self.lm_head.bias = self.lm_head.bias.float()
            self.lm_head.dense.weight = self.lm_head.dense.weight.float()
            self.lm_head.dense.bias = self.lm_head.dense.bias.float()
            self.lm_head.layer_norm.weight = self.lm_head.layer_norm.weight.float()
            self.lm_head.layer_norm.bias = self.lm_head.layer_norm.bias.float()
            x = self.lm_head(x)
        else:
            x = F.linear(x, self.embed_out, bias=self.embed_out_bias)
            x = x.transpose(0, 1)  # (T, B, E) => (B, T, E)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # attentions: B x L x H x T x T
            attentions = torch.stack(attn_weights, 1)
            if self.model_version == "ESM-1":
                # ESM-1 models have an additional null-token for attention, which we remove
                attentions = attentions[..., :-1]
            if padding_mask is not None:
                attention_mask = 1 - padding_mask.type_as(attentions)
                attention_mask = attention_mask.unsqueeze(1) * attention_mask.unsqueeze(2)
                attentions = attentions * attention_mask[:, None, None, :, :]
            result["attentions"] = attentions
            if return_contacts:
                contacts = self.contact_head(tokens, attentions)
                result["contacts"] = contacts

        return result




import numpy as np

from plants_sm.featurization.encoding.encoder import Encoder


class OneHotEncoder(Encoder):
    """
    Generic class to encode sequences with one-hot encoding.

    Parameters
    ----------
    """
    name = "one_hot_encoder"
    output_shape_dimension = 3

    def _set_tokens(self, i: int, token: str):
        one_hot = np.zeros(len(self.alphabet))
        one_hot[i] = 1
        self.tokens[token] = one_hot




from abc import abstractmethod
from typing import List, Union

from pydantic import BaseModel


class Tokenizer(BaseModel):
    """
    Tokenizer class that is used to encode and decode sequences.
    """

    class Config:
        """
        Model Configuration: https://pydantic-docs.helpmanual.io/usage/model_config/
        """
        extra = 'allow'
        allow_mutation = True
        validate_assignment = True
        underscore_attrs_are_private = True

    @abstractmethod
    def tokenize(self, sequence: str) -> List[str]:
        """
        Abstract method that has to be implemented by all tokenizers to tokenize a sequence.
        """





from typing import List

import pandas as pd
from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer


class CSVReader(Reader):
    """
    Class that implements the class Reader and reads CSV, TSV and other format files.
    """

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, sep: str = ",", **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path

        sep : string
            the separator of each column in the file

        Returns
        -------
        """
        super().__init__(filepath_or_buffer, sep=sep, **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the reader can read.

        Returns
        -------
        file_types : List[str]
            object that returns the file types that the reader can read.
        """
        return ['txt', 'csv', 'tsv']

    def read(self) -> pd.DataFrame:
        """
        Method to read the data from file(s) and returns a pandas pd.DataFrame.

        Parameters
        ----------

        Returns
        -------
        data : pd.DataFrame
            Read DataFrame
        """
        return pd.read_csv(self.path, **self.kwargs)


class CSVWriter(Writer):
    """
    Class that implements the class Writer and writes CSV, TSV and other format files.
    """

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, sep: str = ",", **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Writer) and the pd.DataFrame to be written.

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path

        sep : string
            the separator of each column in the file

        Returns
        -------
        """

        super().__init__(filepath_or_buffer, sep=sep, **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the writer can write.

        Returns
        -------
        file_types : List[str]
            the file types that the writer can write.
        """
        return ['txt', 'csv', 'tsv']

    def write(self, df: pd.DataFrame) -> bool:
        """
        Method to write the pd.DataFrame into a CSV file.

        Parameters
        ----------
        df : pd.DataFrame
            pandas DataFrame to be written

        Returns
        -------
        flag : boolean
            whether the pd.DataFrame was written without errors or not
        """
        try:
            df.to_csv(self.buffer, **self.kwargs)
            return True
        except FileNotFoundError:
            return False


def read_csv(filepath_or_buffer: FilePathOrBuffer, **kwargs) -> pd.DataFrame:
    """
    Function that reads the CSV file and returns a pandas pd.DataFrame.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    df : pd.DataFrame
        Read DataFrame
    """

    reader = CSVReader(filepath_or_buffer=filepath_or_buffer, **kwargs)
    df = reader.read()
    return df


def write_csv(filepath_or_buffer: FilePathOrBuffer, df: pd.DataFrame, **kwargs) -> bool:
    """
    Function that writes a pandas pd.DataFrame in a CSV file.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    df : pd.DataFrame
        Dataframe to be written in the CSV file

    Returns
    -------
    flag : boolean
        whether the pd.DataFrame was written without errors
    """

    writer = CSVWriter(filepath_or_buffer=filepath_or_buffer, **kwargs)
    ddf = writer.write(df=df)
    writer.close_buffer()
    return ddf



from plants_sm.featurization.encoding.encoder import Encoder


class LabelEncoder(Encoder):

    output_shape_dimension = 2

    def _set_tokens(self, i: int, token: str):
        self.tokens[token] = i + 1



from typing import List, Union
import json
from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer


class JSONReader(Reader):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path

        Returns
        -------
        """
        super().__init__(filepath_or_buffer, mode="r", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        File types that the reader can read

        Returns
        -------
        file types: List[str]
            List of file types that the reader can read
        """
        return ["json"]

    def read(self) -> Union[dict, List[dict]]:
        """
        Method to read the inputted file or buffer

        Returns
        -------
        data: Union[dict, List[dict]]
            Python object representing the JSON data
        """
        return json.load(self.buffer)


class JSONWriter(Writer):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Writer) and the Python object to be written.

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path

        Returns
        -------
        """

        super().__init__(filepath_or_buffer, **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        File types that the reader can read

        Returns
        -------
        file types: List[str]
            List of file types that the reader can read
        """
        return ["json"]

    def write(self, object_to_be_written: Union[dict, List[dict]]) -> bool:
        """

        Parameters
        ----------
        object_to_be_written: Union[dict, List[dict]]
            Python object representing the JSON data to be written

        Returns
        -------
        flag : boolean
            whether the JSON data was written without errors or not
        """
        try:
            json.dump(object_to_be_written, self.buffer, **self.kwargs)
            return True
        except FileNotFoundError:
            return False


def read_json(file_path: FilePathOrBuffer, **kwargs) -> Union[dict, List[dict]]:
    """
    Function to read JSON file calling the JSONReader method.

    Parameters
    ----------
    file_path: FilePathOrBuffer
        path of the file to be read
    Returns
    -------
    data: Union[dict, List[dict]]
        Python object representing the JSON data
    """
    reader = JSONReader(file_path, **kwargs)
    data = reader.read()
    reader.close_buffer()
    return data


def write_json(output_path: FilePathOrBuffer, object_to_be_written: Union[dict, List[dict]], **kwargs) -> bool:
    """
    Function to write a JSON file calling the JSONWriter method.

    Parameters
    ----------
    object_to_be_written: Union[dict, List[dict]]
        Python object representing the JSON data to be written
    output_path: FilePathOrBuffer
        path or buffer of the file to be written

    Returns
    -------
    flag:bool
        flag that says whether the file was written
    """
    writer = JSONWriter(filepath_or_buffer=output_path, **kwargs)
    flag = writer.write(object_to_be_written)
    writer.close_buffer()
    return flag



import subprocess
from abc import ABC, abstractmethod

import pandas as pd
from tqdm import tqdm


class Alignment(ABC):

    def __init__(self, database):
        self.results = None
        self.database = database

    @abstractmethod
    def create_database(self, fasta_file):
        pass

    def run(self, query_file, output_file, evalue, num_hits, output_options=None, outfmt=6):
        if output_options is None:
            output_options = ["qseqid", "sseqid", "pident", "length", "mismatch",
                              "gapopen", "qstart", "qend", "sstart", "evalue", "bitscore"]

        output_options_str = " ".join(output_options)
        self._run(query_file, output_file, evalue, num_hits, output_options_str, outfmt)
        if outfmt >= 6:
            self.results = pd.read_csv(output_file, sep="\t", header=None, names=output_options)
            self.results.to_csv(output_file, sep="\t", index=False)

    @abstractmethod
    def _run(self, query_file, output_file, evalue, num_hits, output_options_str, outfmt):
        pass

    def associate_to_ec(self, database_ec_dataframe, output_file, columns=None):
        if columns is None:
            columns = ["qseqid",
                       "accession", "pident", "length", "mismatch",
                       "gapopen", "qstart", "qend",
                       "sstart", "evalue", "bitscore"]
        database_reduced = database_ec_dataframe[
            database_ec_dataframe.loc[:, "accession"].isin(self.results.loc[:, "sseqid"])]
        del database_ec_dataframe
        self.results.columns = columns
        database_reduced.drop(columns=["sequence"], inplace=True)
        merged_df = pd.merge(self.results, database_reduced, on='accession', how='inner')
        merged_df.to_csv(output_file, index=False)


class Diamond(Alignment):

    def __init__(self, database):
        super().__init__(database)

    def create_database(self, fasta_file):
        subprocess.run(["diamond", "makedb", "--in", fasta_file, "--db", self.database])

    def _run(self, query_file, output_file, evalue, num_hits, output_options_str, outfmt=6):
        subprocess.call(f"diamond blastp -d {self.database} -q {query_file} -o {output_file} --outfmt {outfmt} "
                        f"{output_options_str} --evalue {evalue} --max-target-seqs {num_hits}", shell=True)


class BLAST(Alignment):

    def __init__(self, database):
        super().__init__(database)

    def create_database(self, fasta_file):
        subprocess.run(["makeblastdb", "-in", fasta_file, "-dbtype", "prot", "-out", self.database])

    def _run(self, query_file, output_file, evalue, num_hits, output_options_str, outfmt=6):
        subprocess.run(["blastp", "-query", query_file, "-db", self.database, "-out", output_file, "-outfmt",
                        f"{outfmt} {output_options_str}", "-evalue", str(evalue), "-max_target_seqs", str(num_hits)])



import math

from ._constants import (AA_ALPHABET,
                         HYDROPHOBICITY_NORMALIZED, HYDROPHILICITY_NORMALIZED, RESIDUE_MASS_NORMALIZED,
                         PK1_NORMALIZED, PK2_NORMALIZED, PI_NORMALIZED)
from ._utils import Descriptor
from .amino_acid_composition import AminoAcidCompositionDescriptor

_AAPS = [HYDROPHOBICITY_NORMALIZED, HYDROPHILICITY_NORMALIZED, RESIDUE_MASS_NORMALIZED,
         PK1_NORMALIZED, PK2_NORMALIZED, PI_NORMALIZED]


def get_correlation_function(ri='S', rj='D'):
    """
    Calculates the correlation function for a given pair of amino acids.

    Parameters
    ----------
    ri
    rj
    """
    theta = 0.0
    for aap in _AAPS:
        theta += math.pow(aap[ri] - aap[rj], 2)

    # normalize by the number of AAPs
    result = round(theta / 6, 3)
    return result


def get_sequence_order_correlation_factor(protein_sequence, k=1):
    """
    Calculates the sequence order correlation factor for a given protein sequence.

    Parameters
    ----------
    protein_sequence
    k
    """
    length_sequence = len(protein_sequence)
    res = []
    for i in range(length_sequence - k):
        aa1 = protein_sequence[i]
        aa2 = protein_sequence[i + k]
        res.append(get_correlation_function(aa1, aa2))
    result = round(sum(res) / (length_sequence - k), 3)
    return result


def get_correlation_function_for_apaac(ri='S', rj='D'):
    """
    Calculates the correlation function for a given pair of amino acids.

    Parameters
    ----------
    ri  : str
        The first amino acid.
    rj  : str
        The second amino acid.
    """

    theta1 = round(HYDROPHOBICITY_NORMALIZED[ri] * HYDROPHOBICITY_NORMALIZED[rj], 3)
    theta2 = round(HYDROPHILICITY_NORMALIZED[ri] * HYDROPHILICITY_NORMALIZED[rj], 3)

    return theta1, theta2


def get_sequence_order_correlation_factor_for_apaac(protein_sequence, k=1):
    """
    Calculates the sequence order correlation factor for a given protein sequence.

    Parameters
    ----------
    protein_sequence
    k
    """
    length_sequence = len(protein_sequence)
    res_hydrophobicity = []
    res_hydrophilicity = []

    for i in range(length_sequence - k):
        aa1 = protein_sequence[i]
        aa2 = protein_sequence[i + k]
        hydrophobicity, hydrophilicity = get_correlation_function_for_apaac(aa1, aa2)
        res_hydrophobicity.append(hydrophobicity)
        res_hydrophilicity.append(hydrophilicity)

    return (round(sum(res_hydrophobicity) / (length_sequence - k), 3),
            round(sum(res_hydrophilicity) / (length_sequence - k), 3))


class PseudoAminoAcidCompositionDescriptor(Descriptor):
    """
    A descriptor that returns the Pseudo Amino Acid Composition of the sequence.
    """

    def __call__(self, sequence: str, lambda_: int = 10, weight: float = 0.05, **kwargs):
        """
        Calculates the Pseudo Amino Acid Composition of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Pseudo Amino Acid Composition for.

        lambda_ : int, default=10
            The lambda value to use for the calculation.

        weight : float, default=0.05
            The weight to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The Pseudo Amino Acid Composition of the sequence.
        """
        correlation_factors = [get_sequence_order_correlation_factor(sequence, i + 1) for i in range(lambda_)]
        correlation_factor = 1 + (weight * sum(correlation_factors))

        aac_counter = AminoAcidCompositionDescriptor()
        aac = aac_counter(sequence)
        paas = [round(aa / correlation_factor, 3) for aa in aac]

        lambdas = [round(weight * correlation_factors[i - 20] / correlation_factor * 100, 3)
                   for i in range(20, 20 + lambda_)]

        return paas + lambdas

    def get_features_out(self, lambda_: int = 10, **kwargs):
        paas = [f'PAAC_{aa}' for aa in AA_ALPHABET]
        lambdas = [f'PAAC_lambda_{i}' for i in range(1, lambda_ + 1)]
        return paas + lambdas


class AmphiphilicPseudoAminoAcidCompositionDescriptor(Descriptor):
    """
    A descriptor that returns the Amphiphilic Pseudo Amino Acid Composition of the sequence.
    """

    def __call__(self, sequence: str, lambda_: int = 10, weight: float = 0.05, **kwargs):
        """
        Calculates the Amphiphilic Pseudo Amino Acid Composition of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Pseudo Amino Acid Composition for.

        lambda_ : int, default=10
            The lambda value to use for the calculation.

        weight : float, default=0.05
            The weight to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The Amphiphilic Pseudo Amino Acid Composition of the sequence.
        """
        correlation_factors = []

        for i in range(lambda_):
            correlation_factors.extend(get_sequence_order_correlation_factor_for_apaac(sequence, i + 1))

        correlation_factor = 1 + (weight * sum(correlation_factors))

        aac_counter = AminoAcidCompositionDescriptor()
        aac = aac_counter(sequence)
        paas = [round(aa / correlation_factor, 3) for aa in aac]

        lambdas = [round(weight * correlation_factors[i - 20] / correlation_factor * 100, 3)
                   for i in range(20, 20 + (2 * lambda_))]

        return paas + lambdas

    def get_features_out(self, lambda_: int = 10, **kwargs):
        paas = [f'APAAC_{aa}' for aa in AA_ALPHABET]
        lambdas = [f'APAAC_lambda_{i}' for i in range(1, (2 * lambda_) + 1)]
        return paas + lambdas



import os
from tempfile import TemporaryDirectory
from typing import Union, List, Tuple
import uuid

from pandas import read_csv

from plants_sm.design_patterns.observer import Observer, Subject
from plants_sm.io import CSVWriter, write_csv, CSVReader
from plants_sm.io.h5 import H5Reader, read_h5, H5Writer, write_h5
from plants_sm.io.json import JSONWriter, write_json, read_json
from plants_sm.io.pickle import PickleReader, PickleWriter, write_pickle, read_pickle


class BatchManager(Observer):

    def __init__(self, batch_size: Union[int, None] = None):
        """
        Class that manages the batches of a dataset.

        Parameters
        ----------
        batch_size : int
            the size of the batches
        """
        self._variables_to_save = None
        self._cls = None
        self.batch_size = batch_size
        filename = str(uuid.uuid4().clock_seq)
        folder = TemporaryDirectory(prefix=filename)
        self.temporary_folder = folder
        self.counter = batch_size
        self.batch_i = 0

    def __del__(self):
        """
        Deletes the temporary folder.
        """
        self.temporary_folder.cleanup()

    def end(self):
        """
        Resets the counter.
        """
        self.counter = 0
        self.batch_i = 0

    def update(self, subject: 'Dataset', **kwargs) -> None:
        """
        Updates the batch manager.

        Parameters
        ----------
        subject: Subject
            the subject that is being observed
        """
        if kwargs["function"] == "__next__":
            self.write_intermediate_files()
            self.counter += self.batch_size

        elif kwargs["function"] == "next_batch":
            if self.counter == 0:
                self.counter += self.batch_size
                subject._batch_state = self.read_intermediate_files(subject)
                if subject._folder_to_load_features is not None and subject._batch_state:
                    self._read_features(subject)
                    self.batch_i += 1

            else:
                self.write_intermediate_files()
                self.counter += self.batch_size
                subject._batch_state = self.read_intermediate_files(subject)
                if subject._folder_to_load_features is not None and subject._batch_state:
                    self._read_features(subject)
                    self.batch_i += 1

    def _read_features(self, subject: 'Dataset'):
        """
        Reads the features from the files.
        """
        subject.features = read_pickle(os.path.join(subject._folder_to_load_features,
                                                    f"features_{self.batch_i}.pkl"))

    def register_class(self, cls, variables_to_save: List[Tuple[str, str]] = None):
        """
        Registers the class that will be used to create the batches.

        Parameters
        ----------
        cls: class
            the class that will be used to create the batches
        variables_to_save: List[Tuple[str, str]]
            the variables that will be saved in the intermediate files
        """
        self._cls = cls
        self._variables_to_save = variables_to_save

    def write_intermediate_files(self):
        """
        Creates the intermediate files to be used in the batches.
        """

        os.makedirs(os.path.join(self.temporary_folder.name, str(self.counter)), exist_ok=True)

        for variable_name, variable_format in self._variables_to_save:
            if variable_format in JSONWriter.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.json")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    write_json(file_path, variable)

            elif variable_format in CSVWriter.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.csv")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    write_csv(file_path, variable, index=False)

            elif variable_format in PickleWriter.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.pkl")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    write_pickle(file_path, variable)

            elif variable_format in H5Writer.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.h5")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    write_h5(variable, file_path)

    def read_intermediate_files(self, subject: 'Dataset') -> bool:
        """
        Reads the intermediate files to be used in the batches.

        Parameters
        ----------
        subject: Subject
            the subject that is being observed

        Returns
        -------
        bool
            True if the files were read, False otherwise
        """

        if not os.path.exists(os.path.join(self.temporary_folder.name, str(self.counter))):
            return False

        for variable_name, variable_format in self._variables_to_save:
            if variable_format in JSONWriter.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.json")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    _variable = read_json(file_path)
                    setattr(subject, variable_name, _variable)
            elif variable_format in CSVReader.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.csv")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    _variable = read_csv(file_path)
                    setattr(subject, variable_name, _variable)
            elif variable_format in PickleReader.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.pkl")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    _variable = read_pickle(file_path)
                    setattr(subject, variable_name, _variable)
            elif variable_format in H5Reader.file_types():
                file_path = os.path.join(self.temporary_folder.name, str(self.counter),
                                         f"{variable_name}.h5")

                variable = getattr(self._cls, variable_name)
                if variable is not None:
                    _variable = read_h5(file_path)
                    setattr(subject, variable_name, _variable)

        return True



from typing import Any, Dict, List, Union, IO, TextIO

import dask
from yaml import CLoader as Loader, safe_load
from yaml import load, dump

from plants_sm.io.commons import get_handle_to_read, get_handle_to_write, FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer


class YAMLReader(Reader):

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the reader can read.

        Returns
        -------
        file_types : List[str]
            file types that the reader can read.
        """
        return ['yml', 'yaml']

    @dask.delayed
    def read(self) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        Method to read the data from file(s) and returns a dictionary or list of dictionaries.

        Parameters
        ----------

        Returns
        -------
        data : Union[List[Dict[str, Any]], Dict[str, Any]]
        """
        if 'Loader' not in self.kwargs:
            self.kwargs['Loader'] = Loader

        mode = self.kwargs.pop('mode', 'r')
        auth = self.kwargs.pop('auth', None)
        if "preserve_order" in self.kwargs:
            preserve_order = self.kwargs.pop("preserve_order")
        else:
            preserve_order = False
        handle = get_handle_to_read(self.buffer, mode=mode, auth=auth)

        with handle as f:
            if preserve_order:
                return load(f, **self.kwargs)
            else:
                return safe_load(f)


class YAMLWriter(Writer):

    @staticmethod
    def file_types() -> List[str]:
        """
        Returns the file types that the writer can write.

        Returns
        -------
        file_types : List[str]
            file types that the writer can write.
        """
        return ['yml', 'yaml']

    def write(self, data: Union[List[Dict[str, Any]], Dict[str, Any]]) -> bool:
        """
        Method to write the dictionary into a YAML file.

        Parameters
        ----------
        data : Any
            object to be written (e.g., dataframe or data file)

        Returns
        -------
        flag : boolean
            whether the DataFrame was written without errors or not
        """
        mode = self.kwargs.pop('mode', 'w')
        handle = get_handle_to_write(self.buffer, mode=mode)
        with handle as f:
            dump(data, f, **self.kwargs)
        return True


def read_yaml(file_path_or_buffer: FilePathOrBuffer, **kwargs) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    """
    Function that reads the YAML file and returns a dictionary.

    Parameters
    ----------
    file_path_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    data : Union[List[Dict[str, Any]], Dict[str, Any]]
        dictionary with the information read from the YAML file
    """
    with YAMLReader(file_path_or_buffer=file_path_or_buffer, **kwargs) as reader:
        # suppress warning due to dask delayed decorator: https://github.com/dask/dask/issues/7779
        # noinspection PyUnresolvedReferences
        data = reader.read().compute()
    return data


def write_yaml(file_path_or_buffer: FilePathOrBuffer,
               data: Union[List[Dict[str, Any]], Dict[str, Any]],
               **kwargs) -> bool:
    """
    Function that writes a dictionary in a YAMl file.

    Parameters
    ----------
    file_path_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path
    data : Union[List[Dict[str, Any]], Dict[str, Any]]
        object with the information to be written

    Returns
    -------
    flag : boolean
        whether the DataFrame was written without errors
    """
    with YAMLWriter(file_path_or_buffer=file_path_or_buffer, **kwargs) as writer:
        flag = writer.write(data=data)
    return flag



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pandas import DataFrame
from sklearn.decomposition import PCA
import seaborn as sns

from plants_sm.unsupervised.model import UnsupervisedModel


class SklearnPCA(UnsupervisedModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pca = PCA(**self.kwargs)

    def _predict(self, x):
        pass

    def _fit(self, x):
        self.pca.fit(x)
        self.fitted = True
        return self

    def _transform(self, x) -> pd.DataFrame:
        if self.fitted:
            return self.pca.transform(x)
        else:
            raise ValueError("The model is not fitted")

    def generate_pca_variance_plot(self):
        plt.figure()
        plt.plot(range(self.pca.n_components_), np.cumsum(self.pca.explained_variance_ratio_ * 100))
        for i in range(self.pca.n_components_):
            cumulative_sum = np.sum(self.pca.explained_variance_ratio_[:i])
            if cumulative_sum > 0.95:
                break
        plt.xlabel('Number of Components')
        plt.ylabel('Variance (%)')
        plt.title('Cumulative Explained Variance')
        plt.xlim(0, self.pca.n_components_)
        plt.ylim(30, 100)
        plt.axhline(y=95, color='r')
        plt.axvline(x=i, color='g')
        plt.show()
        print(f'First 2 PC: {sum(self.pca.explained_variance_ratio_[0:2] * 100)}')
        print(f'First {i} PC: {sum(self.pca.explained_variance_ratio_[0:i] * 100)}')

    def generate_dotplot(self, data, labels):
        plt.figure(figsize=(20, 20))
        plt.xlabel(f'PC1 = {np.round(self.pca.explained_variance_ratio_[0] * 100, 3)}% variance')
        plt.ylabel(f'PC2 = {np.round(self.pca.explained_variance_ratio_[1] * 100, 3)}% variance')

        data = DataFrame(data)
        data["label"] = labels
        sns.scatterplot(
            data=data,
            x=data.columns[0], y=data.columns[1],
            hue="label",
            palette=sns.color_palette("deep", len(np.unique(data["label"]))),
            legend="full",
            s=25)



from __future__ import annotations
from abc import ABC, abstractmethod


class Subject(ABC):
    """
    The Subject interface declares a set of methods for managing subscribers.
    """

    @abstractmethod
    def attach(self, observer: Observer) -> None:
        """
        Attach an observer to the subject.
        """

    @abstractmethod
    def detach(self, observer: Observer) -> None:
        """
        Detach an observer from the subject.
        """

    @abstractmethod
    def notify(self) -> None:
        """
        Notify all observers about an event.
        """


class ConcreteSubject(Subject):
    """
    The Subject owns some important state and notifies observers when the state
    changes.
    """

    _state: int = None
    """
    For the sake of simplicity, the Subject's state, essential to all
    subscribers, is stored in this variable.
    """

    _observer: Observer = None
    """
    List of subscribers. In real life, the list of subscribers can be stored
    more comprehensively (categorized by event type, etc.).
    """

    def attach(self, observer: Observer) -> None:
        self._observer = observer

    def detach(self, observer: Observer) -> None:
        self._observer = None

    """
    The subscription management methods.
    """

    def notify(self, **kwargs) -> None:
        """
        Trigger an update in each subscriber.
        """

        self._observer.update(self, **kwargs)

    def end(self):
        self._observer.end()


class Observer(ABC):
    """
    The Observer interface declares the update method, used by subjects.
    """

    @abstractmethod
    def update(self, subject: Subject, **kwargs) -> None:
        """
        Receive update from subject.
        """

    @abstractmethod
    def end(self):
        """
        End the observer.
        """




from typing import Tuple
from plants_sm.data_structures.dataset.dataset import Dataset
from plants_sm.transformation.transformer import Transformer

from plants_sm.transformation._utils import transform_instances


class Truncator(Transformer):

    max_length: int

    def _fit(self, dataset: Dataset, instance_type: str) -> 'Truncator':
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            Dataset to fit.

        Returns
        -------
        self : Truncator
            Fitted transformer.
        """
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'Truncator':
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            Dataset to fit.

        Returns
        -------
        self : Truncator
            Fitted transformer.
        """
        return self._fit(dataset, instance_type)

    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        Standardize protein sequences.

        Parameters
        ----------
        dataset : Dataset
            Dataset to transform.

        Returns
        -------
        dataset : Dataset
            Transformed dataset.
        """
        return transform_instances(self.n_jobs, dataset, self._truncate, instance_type,
                                   self.__class__.__name__)

    def _truncate(self, sequence: str, identifier: str) -> Tuple[str, str]:
        """
        Preprocess a protein sequence.

        Parameters
        ----------
        sequence : str
            Protein sequence to preprocess.
        identifier : str
            Protein identifier.

        Returns
        -------
        Tuple[str, str]
            Tuple with the identifier as first value and the preprocessed sequence as second.
        """
        if len(sequence) > self.max_length:
            sequence = sequence[:self.max_length]
        return identifier, sequence 






# Supported problem types
from enum import Enum

BINARY = 'binary'
MULTICLASS = 'multiclass'
REGRESSION = 'regression'
SOFTCLASS = 'softclass'
QUANTILE = 'quantile'

PROBLEM_TYPES = [BINARY, MULTICLASS, REGRESSION, SOFTCLASS, QUANTILE]


# files constants

class FileConstants(Enum):

    PYTORCH_MODEL_WEIGHTS = 'pytorch_model_weights.pt'
    PYTORCH_MODEL_PKL = 'pytorch_model.pkl'

    TENSORFLOW_MODEL = 'tensorflow_model.h5'

    MODEL_PARAMETERS_PKL = 'model_parameters.pkl'

    MODEL_CLASSES_FILES_PKL = [PYTORCH_MODEL_PKL, TENSORFLOW_MODEL]


import os
import time

import numpy as np
import pandas as pd
from plants_sm.models.constants import FileConstants
import torch
from torch import nn
from tqdm import tqdm

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io.pickle import read_pickle, write_pickle
from plants_sm.models.ec_number_prediction._clean_distance_maps import compute_esm_distance, get_dist_map, \
    get_ec_id_dict, model_embedding_test, get_dist_map_test, random_nk_model, get_random_nk_dist_map
from plants_sm.models.ec_number_prediction._clean_utils import SupConHardLoss, get_dataloader
from plants_sm.models.model import Model
from plants_sm.models.ec_number_prediction._clean_distance_maps import divide_labels_by_EC_level

from plants_sm.models._utils import write_model_parameters_to_pickle

import logging
import os
from logging.handlers import TimedRotatingFileHandler


class LayerNormNet(nn.Module):
    def __init__(self, input_dim, hidden_dim, out_dim, device, dtype, drop_out=0.1):
        super(LayerNormNet, self).__init__()
        self.hidden_dim1 = hidden_dim
        self.out_dim = out_dim
        self.drop_out = drop_out
        self.device = device
        self.dtype = dtype
        self.input_dim = input_dim

        self.fc1 = nn.Linear(input_dim, hidden_dim, dtype=dtype, device=device)
        self.ln1 = nn.LayerNorm(hidden_dim, dtype=dtype, device=device)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim,
                             dtype=dtype, device=device)
        self.ln2 = nn.LayerNorm(hidden_dim, dtype=dtype, device=device)
        self.fc3 = nn.Linear(hidden_dim, out_dim, dtype=dtype, device=device)
        self.dropout = nn.Dropout(p=drop_out)

    def forward(self, x):
        x = self.dropout(self.ln1(self.fc1(x)))
        x = torch.relu(x)
        x = self.dropout(self.ln2(self.fc2(x)))
        x = torch.relu(x)
        x = self.fc3(x)
        return x


class PValueInference:
    def __init__(self, p_value=1e-5, nk_random=20):
        self.p_value = p_value
        self.nk_random = nk_random


class CLEANSupConH(Model):

    def __init__(self,
                 distance_map_path: str, input_dim, hidden_dim=512, out_dim=256, dtype=torch.float32,
                 device: str = "cuda", drop_out=0.1,
                 lr=5e-4, epochs=1500, n_pos=9, n_neg=30, adaptative_rate=10,
                 temp=0.1, batch_size=6000, verbose=True, ec_label="EC",
                 model_name="CLEANSupConH", path_to_save_model="./data/model/",
                 evaluation_class=PValueInference(p_value=1e-5, nk_random=20), logger_path="clean_supconh.log"):
        self.distance_map_path = distance_map_path
        parent_folder = os.path.dirname(distance_map_path)
        if not os.path.exists(parent_folder):
            os.makedirs(parent_folder)

        self.path_to_save_model = path_to_save_model
        if not os.path.exists(self.path_to_save_model):
            os.makedirs(self.path_to_save_model)

        self.device = device
        self.model = LayerNormNet(input_dim, hidden_dim, out_dim, device, dtype, drop_out)
        self.dtype = dtype
        self._history = {"train_loss": []}
        self.lr = lr
        self.epochs = epochs
        self.n_pos = n_pos
        self.n_neg = n_neg
        self.adaptative_rate = adaptative_rate
        self.temp = temp
        self.batch_size = batch_size
        self.loss_function = SupConHardLoss
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.lr, betas=(0.9, 0.999))
        self.verbose = verbose
        self.ec_label = ec_label
        self.model_name = model_name
        self.evaluation_class = evaluation_class

        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        if logger_path:
            handler = TimedRotatingFileHandler(logger_path, when='midnight', backupCount=30)
        else:
            handler = TimedRotatingFileHandler(f'./{self.model_name}.log', when='midnight', backupCount=20)
        handler.setFormatter(formatter)

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.DEBUG)

        self.rand_nk_ids = None
        self.rand_nk_emb_train = None
        self.random_nk_dist_map = None
    
    def __name__(self):
        return "CLEANSupConH"

    def _preprocess_data(self, dataset: Dataset, **kwargs) -> Dataset:
        pass

    def _fit_data(self, train_dataset: Dataset, validation_dataset: Dataset):
        self.labels = train_dataset._labels_names
        if not os.path.exists(self.distance_map_path + ".pkl") \
                and not os.path.exists(self.distance_map_path + "_esm.pkl"):
            compute_esm_distance(train_dataset, self.distance_map_path, self.device)

        best_loss = float('inf')
        # ======================== override args ====================#
        self.logger.info(f'==> device used: {self.device} | dtype used: {self.dtype}\n==> args:')
        # ======================== ESM embedding  ===================#
        # loading ESM embedding for dist map

        self.logger.info(f'==> loading ESM embedding for dist map')
        self._esm_emb_train = read_pickle(self.distance_map_path + '_esm.pkl').to(device=self.device, dtype=self.dtype)
        dist_map = read_pickle(self.distance_map_path + '.pkl')

        self._id_ec_train, self._ec_id_dict_train = get_ec_id_dict(train_dataset, self.ec_label)
        ec_id = {key: list(self._ec_id_dict_train[key]) for key in self._ec_id_dict_train.keys()}
        # ======================== initialize model =================#
        train_loader = get_dataloader(dist_map, self._id_ec_train, ec_id, self.n_pos, self.n_neg, train_dataset,
                                      self.batch_size)
        
        self.logger.info(f"The number of unique EC numbers: {len(dist_map.keys())}")

        # ======================== training =======-=================#
        # training
        for epoch in range(1, self.epochs + 1):
            if epoch % self.adaptative_rate == 0 and epoch != self.epochs + 1:
                # save updated model
                torch.save(self.model.state_dict(), os.path.join(self.path_to_save_model,
                                                                 self.model_name + '_' + str(epoch) + '.pth'))
                # delete last model checkpoint
                if epoch != self.adaptative_rate:
                    os.remove(os.path.join(self.path_to_save_model, self.model_name + '_' +
                                           str(epoch - self.adaptative_rate) + '.pth'))
                # sample new distance map
                dist_map = get_dist_map(
                    self._ec_id_dict_train, self._esm_emb_train, self.device, self.dtype, model=self.model)
                train_loader = get_dataloader(dist_map, self._id_ec_train, ec_id, self.n_pos, self.n_neg, train_dataset,
                                              self.batch_size)
            # -------------------------------------------------------------------- #
            epoch_start_time = time.time()
            train_loss = self._train(train_loader, epoch)
            # only save the current best model near the end of training
            if train_loss < best_loss and epoch > 0.8 * self.epochs:
                torch.save(self.model.state_dict(), os.path.join(self.path_to_save_model, self.model_name + '.pth'))
                best_loss = train_loss
                self.logger.info(f'Best from epoch : {epoch:3d}; loss: {train_loss:6.4f}')

            elapsed = time.time() - epoch_start_time
            self.logger.info('-' * 75)
            self.logger.info(f'| end of epoch {epoch:3d} | time: {elapsed:5.2f}s | '
                  f'training loss {train_loss:6.4f}')
            self.logger.info('-' * 75)
            if "loss" not in self._history:
                self._history["loss"] = []
            self._history["loss"].append(train_loss)

            # emb_train = self.model(self._esm_emb_train)

            # self.rand_nk_ids, self.rand_nk_emb_train = random_nk_model(
            #     self._id_ec_train, self._ec_id_dict_train, emb_train, n=self.evaluation_class.nk_random, weighted=True)

            # self.random_nk_dist_map = get_random_nk_dist_map(
            #     emb_train, self.rand_nk_emb_train, self._ec_id_dict_train, self.rand_nk_ids, self.device, self.dtype)
            
            # predictions = self._predict(validation_dataset)
            # self._history["validation_loss"] = self.evaluation_class.evaluate(validation_dataset, predictions)
            
        # remove tmp save weights
        if os.path.exists(os.path.join(self.path_to_save_model, self.model_name + '.pth')):
            os.remove(os.path.join(self.path_to_save_model, self.model_name + '.pth'))
        if os.path.exists(os.path.join(self.path_to_save_model, self.model_name + '_' + str(epoch) + '.pth')):
            os.remove(os.path.join(self.path_to_save_model, self.model_name + '_' + str(epoch) + '.pth'))
        # save final weights
        torch.save(self.model.state_dict(), os.path.join(self.path_to_save_model, self.model_name + '.pth'))

    def _train(self, train_loader, epoch):
        self.model.train()
        total_loss = 0.
        start_time = time.time()

        for batch, data in enumerate(train_loader):
            self.optimizer.zero_grad()
            model_emb = self.model(data.to(device=self.device, dtype=self.dtype))
            loss = self.loss_function(model_emb, self.temp, self.n_pos)
            loss.backward()
            self.optimizer.step()

            total_loss += loss.item()
            if self.verbose:
                ms_per_batch = (time.time() - start_time) * 1000
                cur_loss = total_loss
                self.logger.info(f'| epoch {epoch:3d} | {batch:5d}/{len(train_loader):5d} batches | '
                      f'lr {self.lr:02.4f} | ms/batch {ms_per_batch:6.4f} | '
                      f'loss {cur_loss:5.2f}')
                start_time = time.time()
            self.logger.info(f"batch {batch} done")
        # record running average training loss
        return total_loss / (batch + 1)

    def _predict_proba(self, dataset: Dataset) -> np.ndarray:
        return self._predict(dataset)

    @staticmethod
    def _get_pvalue_choices(df, random_nk_dist_map, p_value=1e-5):
        all_test_EC = set()
        nk = len(random_nk_dist_map.keys())
        threshold = p_value * nk
        ecs = []
        for col in tqdm(df.columns):
            ec = []
            smallest_10_dist_df = df[col].nsmallest(10)
            for i in range(10):
                EC_i = smallest_10_dist_df.index[i]
                # find all the distances in the random nk w.r.t. EC_i
                # then sorted the nk distances
                rand_nk_dists = [random_nk_dist_map[rand_nk_id][EC_i]
                                 for rand_nk_id in random_nk_dist_map.keys()]
                rand_nk_dists = np.sort(rand_nk_dists)
                # rank dist_i among rand_nk_dists
                dist_i = smallest_10_dist_df[i]
                rank = np.searchsorted(rand_nk_dists, dist_i)
                if (rank <= threshold) or (i == 0):
                    dist_str = "{:.4f}".format(dist_i)
                    all_test_EC.add(EC_i)
                    ec.append('EC:' + str(EC_i) + '/' + dist_str)
                else:
                    break
            ec.insert(0, col)
            ecs.append(ec)
        return ecs
    
    def get_final_labels(self, EC1, EC2, EC3, EC4):
        res = np.zeros((len(self.labels),), dtype=object)
        
        for ec in EC1 + EC2 + EC3 + EC4:
            if ec in self.labels:
                index_ = self.labels.index(ec)
                res[index_] = 1

        return res

    def _predict(self, dataset: Dataset) -> np.ndarray:
        id_ec_test, _ = get_ec_id_dict(dataset, self.ec_label)
        # load precomputed EC cluster center embeddings if possible
        emb_train = self.model(self._esm_emb_train)

        if self.rand_nk_ids is None or self.rand_nk_emb_train is None or self.random_nk_dist_map is None:
            self.rand_nk_ids, self.rand_nk_emb_train = random_nk_model(
                self._id_ec_train, self._ec_id_dict_train, emb_train, n=self.evaluation_class.nk_random, weighted=True)

            self.random_nk_dist_map = get_random_nk_dist_map(
                emb_train, self.rand_nk_emb_train, self._ec_id_dict_train, self.rand_nk_ids, self.device, self.dtype)

        emb_test = model_embedding_test(dataset, id_ec_test, self.model, self.device, self.dtype)

        eval_dist = get_dist_map_test(emb_train, emb_test, self._ec_id_dict_train, id_ec_test,
                                      self.device, self.dtype)

        eval_df = pd.DataFrame.from_dict(eval_dist)

        ecs = self._get_pvalue_choices(eval_df, self.random_nk_dist_map, p_value=self.evaluation_class.p_value)
        
        pred_label = []
        for row in ecs:
            preds_ec_lst = []
            preds_with_dist = row[1:]
            for pred_ec_dist in preds_with_dist:
                # get EC number 3.5.2.6 from EC:3.5.2.6/10.8359
                ec_i = pred_ec_dist.split(":")[1].split("/")[0]
                preds_ec_lst.append(ec_i)
            pred_label.append(";".join(preds_ec_lst))

        res = np.zeros((dataset.dataframe.shape[0], len(self.labels)), )
        for i, ec in enumerate(pred_label):
            
            EC1, EC2, EC3, EC4 = divide_labels_by_EC_level(ec)
            res[i, :] = self.get_final_labels(EC1, EC2, EC3, EC4)

        return res

    @staticmethod
    def _save_pytorch_model(model, path):
        weights_path = os.path.join(path, FileConstants.PYTORCH_MODEL_WEIGHTS.value)
        torch.save(model.state_dict(), weights_path)
        write_pickle(os.path.join(path, FileConstants.PYTORCH_MODEL_PKL.value), model)


    def _save(self, path: str):
        self._save_pytorch_model(self.model, path)

        del self.rand_nk_ids
        del self.rand_nk_emb_train
        del self.random_nk_dist_map

        self.rand_nk_ids = None
        self.rand_nk_emb_train = None
        self.random_nk_dist_map = None

        write_pickle(os.path.join(path, "model_class.pkl"), self)

    @staticmethod
    def _read_model(path):
        """
        Read the model from the specified path.

        Parameters
        ----------
        path: str
            Path to read the model from

        Returns
        -------
        torch.nn.Module
        """
        weights_path = os.path.join(path, FileConstants.PYTORCH_MODEL_WEIGHTS.value)
        model = read_pickle(os.path.join(path, FileConstants.PYTORCH_MODEL_PKL.value))
        model.load_state_dict(torch.load(weights_path))
        model.eval()
        return model

    @classmethod
    def _load(cls, path: str):
        new_class = read_pickle(os.path.join(path, "model_class.pkl"))
        model = cls._read_model(path)
        new_class.model = model

        return new_class

    @property
    def history(self):
        return self._history



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pandas import DataFrame
from sklearn.manifold import TSNE
import seaborn as sns

from plants_sm.unsupervised.model import UnsupervisedModel


class SklearnTSNE(UnsupervisedModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tsne = TSNE(**self.kwargs)

    def _fit(self, x):
        self.tsne.fit(x)
        self.fitted = True
        return self

    def fit_transform(self, x):
        return self.tsne.fit_transform(x)

    def _transform(self, x):
        pass

    def _predict(self, x):
        pass

    @staticmethod
    def generate_dotplot(x, labels=None):
        plt.figure(figsize=(20, 20))
        plt.xlabel('t-SNE 1')
        plt.ylabel('t-SNE 2')
        plot = sns.scatterplot(
            x[:, 0], x[:, 1],
            hue=None,
            palette=sns.color_palette("deep", 2),
            legend="full",
            s=25)
        plt.title("TSNE")

        data = DataFrame(x)
        data["label"] = labels
        sns.scatterplot(
            data=data,
            x=data.columns[0], y=data.columns[1],
            hue="label",
            palette=sns.color_palette("deep", len(np.unique(data["label"]))),
            legend="full",
            s=25)
        return plot



from collections import Counter
from statistics import stdev

from modlamp.descriptors import PeptideDescriptor

from ._constants import (COR_HYDROPHOBICITY_NORMALIZED,
                         COR_AV_FLEXIBILITY_NORMALIZED,
                         COR_POLARIZABILITY_NORMALIZED,
                         COR_FREE_ENERGY_NORMALIZED,
                         COR_RESIDUE_ASA_NORMALIZED,
                         COR_RESIDUE_VOL_NORMALIZED,
                         COR_STERIC_NORMALIZED,
                         COR_MUTABILITY_NORMALIZED)
from ._utils import Descriptor

_AAPS = [COR_HYDROPHOBICITY_NORMALIZED,
         COR_AV_FLEXIBILITY_NORMALIZED,
         COR_POLARIZABILITY_NORMALIZED,
         COR_FREE_ENERGY_NORMALIZED,
         COR_RESIDUE_ASA_NORMALIZED,
         COR_RESIDUE_VOL_NORMALIZED,
         COR_STERIC_NORMALIZED,
         COR_MUTABILITY_NORMALIZED]


_AAPS_NAMES = ['COR_HYDROPHOBICITY', 'COR_AV_FLEXIBILITY', 'COR_POLARIZABILITY', 'COR_FREE_ENERGY', 'COR_RESIDUE_ASA',
               'COR_RESIDUE_VOL', 'COR_STERIC', 'COR_MUTABILITY']


def calculate_each_normalized_moreau_broto_auto(protein_sequence, aap):
    result = []
    for i in range(1, 31):
        temp = 0

        for j in range(len(protein_sequence) - i):
            temp += aap[protein_sequence[j]] * aap[protein_sequence[j + 1]]

        if len(protein_sequence) - i == 0:
            result.append(round(temp / (len(protein_sequence)), 3))

        else:
            result.append(round(temp / (len(protein_sequence) - i), 3))

    return result


def calculate_each_moran_auto(protein_sequence, aap):
    counts = Counter(protein_sequence)
    cds = sum([cnt * aap[aa] for aa, cnt in counts.items()])
    pmean = cds / len(protein_sequence)

    cc = [aap[aa] for aa in protein_sequence]
    k = (stdev(cc)) ** 2

    result = []
    for i in range(1, 31):
        temp = 0

        for j in range(len(protein_sequence) - i):
            temp += (aap[protein_sequence[j]] - pmean) * (aap[protein_sequence[j + i]] - pmean)

        if len(protein_sequence) - i == 0:
            result.append(round(temp / (len(protein_sequence)) / k, 3))

        else:
            result.append(round(temp / (len(protein_sequence) - i) / k, 3))

    return result


def calculate_each_geary_auto(protein_sequence, aap):

    cc = [aap[aa] for aa in protein_sequence]

    k = (((stdev(cc)) ** 2) * len(protein_sequence)) / (len(protein_sequence) - 1)

    result = []
    for i in range(1, 31):
        temp = 0

        for j in range(len(protein_sequence) - i):
            temp += (aap[protein_sequence[j]] - aap[protein_sequence[j + i]]) ** 2

        if len(protein_sequence) - i == 0:
            result.append(round(temp / (2 * (len(protein_sequence))) / k, 3))

        else:
            result.append(round(temp / (2 * (len(protein_sequence) - i)) / k, 3))

    return result


class NormalizedMoreauBrotoDescriptor(Descriptor):
    """
    A descriptor that returns the normalized Moreau Broto of the sequence.
    """

    def __call__(self, sequence: str, lambda_: int = 10, weight: float = 0.05, **kwargs):
        """
        Calculates the normalized Moreau Broto of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the normalized Moreau Broto for.

        lambda_ : int, default=10
            The lambda value to use for the calculation.

        weight : float, default=0.05
            The weight to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The normalized Moreau Broto of the sequence.
        """
        result = []
        for aap in _AAPS:
            result.extend(calculate_each_normalized_moreau_broto_auto(sequence, aap))
        return result

    def get_features_out(self, **kwargs):
        return [f'MoreauBrotoAuto_{aap}_{i}' for aap in _AAPS_NAMES for i in range(1, 31)]


class MoranAutoDescriptor(Descriptor):
    """
    A descriptor that returns the Moran Auto of the sequence.
    """

    def __call__(self, sequence: str, lambda_: int = 10, weight: float = 0.05, **kwargs):
        """
        Calculates the Moran Auto of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Moran Auto for.

        lambda_ : int, default=10
            The lambda value to use for the calculation.

        weight : float, default=0.05
            The weight to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The Moran Auto of the sequence.
        """
        result = []
        for aap in _AAPS:
            result.extend(calculate_each_moran_auto(sequence, aap))
        return result

    def get_features_out(self, **kwargs):
        return [f'MoranAuto_{aap}_{i}' for aap in _AAPS_NAMES for i in range(1, 31)]


class GearyAutoDescriptor(Descriptor):
    """
    A descriptor that returns the Geary Auto of the sequence.
    """

    def __call__(self, sequence: str, lambda_: int = 10, weight: float = 0.05, **kwargs):
        """
        Calculates the Geary Auto of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Geary Auto for.

        lambda_ : int, default=10
            The lambda value to use for the calculation.

        weight : float, default=0.05
            The weight to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The Geary Auto of the sequence.
        """
        result = []
        for aap in _AAPS:
            result.extend(calculate_each_geary_auto(sequence, aap))
        return result

    def get_features_out(self, **kwargs):
        return [f'GearyAuto_{aap}_{i}' for aap in _AAPS_NAMES for i in range(1, 31)]


class ModlAMPAutoCorrelationDescriptor(Descriptor):
    """
    A descriptor that returns the Auto correlation of the sequence inferred with modlAMP.
    """
    def __call__(self, sequence: str, window: int = 7, scalename: str = 'Eisenberg', **kwargs):
        """
        Calculates the Auto correlation of the sequence inferred with modlAMP.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Auto correlation inferred with modlAMP for.

        window : int, default=7
            The window to use for the calculation.

        scalename : str, default='Eisenberg'
            The scalename to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The the Auto correlation of the sequence inferred with modlAMP.
        """
        amp = PeptideDescriptor(sequence, scalename)
        amp.calculate_autocorr(window)
        return list(amp.descriptor[0])

    def get_features_out(self, window: int = 7, **kwargs):
        # Auto correlation yields window n values for Eisenberg scalename
        return [f'modlAMP_auto_correlation_{i}' for i in range(1, window + 1)]


class ModlAMPCrossCorrelationDescriptor(Descriptor):
    """
    A descriptor that returns the Cross correlation of the sequence inferred with modlAMP.
    """
    def __call__(self, sequence: str, window: int = 7, scalename: str = 'Eisenberg', **kwargs):
        """
        Calculates the Cross correlation of the sequence inferred with modlAMP.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Cross correlation inferred with modlAMP for.

        window : int, default=7
            The window to use for the calculation.

        scalename : str, default='Eisenberg'
            The scalename to use for the calculation.

        kwargs

        Returns
        -------
        list of float
            The the Auto correlation of the sequence inferred with modlAMP.
        """
        amp = PeptideDescriptor(sequence, scalename)
        amp.calculate_crosscorr(window)
        return list(amp.descriptor[0])

    def get_features_out(self, window: int = 7, **kwargs):
        # Cross correlation yields window n values for Eisenberg scalename
        return [f'modlAMP_cross_correlation_{i}' for i in range(1, window + 1)]
